/*
 * This file is part of Splashy.
 * 
 * Splashy is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any
 * later version.
 * 
 * Splashy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with Splashy; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301 USA 
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifndef _GNU_SOURCE
/*
 * Allow GNU specific funtions: *_unlocked() in stdio.h 
 */
#define _GNU_SOURCE
#endif /* _GNU_SOURCE */

#include <fcntl.h>              /* 0_* and other flags for g_open() */
#include <glib.h>
#include <glib/gstdio.h>        /* g_open() other POSIX flags */
#include <unistd.h>             /* close() fork() getpid() unlink() */

#include <linux/vt.h>           /* VT_* */
#include <sys/ioctl.h>          /* ioctl() */
#include <sys/types.h>          /* needed by regex.h regcomp() */
#include <regex.h>              /* regexec() regcomp() */

#include "common_macros.h"
#include "common_functions.h"
#include "xml_parser.h"
#include "xml_config.h"

/**
 * @desc allows switching virtual terminals
 */
void
splashy_chvt (gint vt_no)
{
        gint fd = g_open ("/dev/tty0", O_RDONLY|O_WRONLY);
        if (fd < 0)
                return;
        if (!ioctl (fd, VT_ACTIVATE, vt_no))
                ioctl (fd, VT_WAITACTIVE, vt_no);
        close (fd);
}

/**
 * @desc takes care of initilizing all special files needed
 * by splashy
 */
gboolean
splashy_init (gboolean create_fifo)
{
        const gchar *_spl_fifo_path=NULL;
        struct stat _stat;

        DEBUG_PRINT ("Entering spl_init(%s)", "");
        if (!init_config (SPL_CONFIG_FILE))
                return FALSE;
        if ( create_fifo )
        {
                _spl_fifo_path = xml_parser_get_text (SPL_FIFO_XPATH);
                /*
                 * re-Create the splash FIFO 
                 */
                g_stat (_spl_fifo_path, &_stat);
                if (!S_ISFIFO (_stat.st_mode))
                {
                        DEBUG_PRINT ("spl_init() attempting to create fifo %s",
                                        _spl_fifo_path);
                        if (mkfifo (_spl_fifo_path, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP) == 0)
                        {
                                return TRUE;
                        }
                        else
                        {
                                ERROR_PRINT ("FIFO creation failed %s",
                                                _spl_fifo_path);
                                return FALSE;   /* assumes that it will fail */
                        }
                }
                DEBUG_PRINT ("Reusing FIFO %s", _spl_fifo_path);
        }
        return TRUE;            /* assumes file exists */
}

inline gboolean
splashy_test_env_pattern (const gchar * perr_pattern, const gchar * filename)
{
        regex_t preg;
        gint max_buf = 1024;    /* we need at least this number of characters 
                                 * to check */
        gboolean ret = FALSE;   /* assumes we will fail */
        if (!g_file_test (filename, G_FILE_TEST_EXISTS))
        {
                ERROR_PRINT ("Cannot test file %s\n", filename);
                return FALSE;
        }
        FILE *fp;
        gchar *buffer = g_try_malloc (max_buf);
        if (buffer && (fp = g_fopen (filename, "r")))
        {
                fread_unlocked (buffer, max_buf, 1, fp);        /* *_unlocked 
                                                                 * are a GNU
                                                                 * extension: 
                                                                 * makes
                                                                 * function
                                                                 * fread()
                                                                 * thread
                                                                 * safe */
                fclose (fp);
                buffer[max_buf - 1] = '\0';
                if (regcomp (&preg, perr_pattern,
                             REG_EXTENDED | REG_ICASE | REG_NEWLINE |
                             REG_NOSUB) == 0)
                {
                        if (regexec (&preg, buffer, 0, NULL, 0) == 0)
                        {
                                ret = TRUE;
                        }
                        else
                        {
                                g_printerr
                                        ("WARN: %s was not found in %s. Did you append vga=0x317 to your kernel parameters? See README for more\n",
                                         perr_pattern, filename);
                        }
                        regfree (&preg);
                }
        }
        if (buffer)
                g_free (buffer);
        return ret;
}
/**
 * @desc looks for pattern in string ignoring known failure text messages
 * @param perr_pattern a valid regular expression pattern. @see man regex
 * @param str string to search for pattern in
 * @return true if pattern was found
 */
inline gboolean
splashy_search_pattern_str (const gchar * perr_pattern, const gchar * str, gshort use_ignore)
{
        gboolean ret = FALSE;   /* assumes we will fail */
        gshort ignore_error = 0; /* we assume that we will be ignoring patterns */
        /* by default we ignore our own errors and loading kernel modules */
        const gchar* ignore_pattern = "(splashy|Cannot[[:space:]]+test[[:space:]]+file|load[[:space:]]+module|fatal:[[:space:]]+module[[:space:]].*[[:space:]]not[[:space:]]found|fatal:[[:space:]]+error[[:space:]]+inserting[[:space:]].*|warning:[[:space:]]+error[[:space:]]+inserting.*)";
        regex_t ignore_preg;
        regex_t preg;

        if (perr_pattern == NULL || str == NULL)
                return FALSE;

        /* ignore_error should be zero */
        if (use_ignore)
        {
                ignore_error = regcomp (&ignore_preg, ignore_pattern,
                                REG_EXTENDED | REG_ICASE | REG_NEWLINE |
                                REG_NOSUB);

                /* 
                 * make sure we are not ignoring this type of errors before
                 * we continue 
                 */
                if ( !ignore_error )
                {
                        if (regexec (&ignore_preg, str, 0, NULL, 0) == 0)
                        {
                                /* we ignore this and return FALSE */
                                regfree (&ignore_preg);
                                goto end;
                        }
                        regfree (&ignore_preg);
                } else {
                        DEBUG_PRINT ("There was an error compiling the ignore_pattern regex: %s",ignore_pattern);
                }
        } 
        
        if (regcomp (&preg, perr_pattern,
                                REG_EXTENDED | REG_ICASE | REG_NEWLINE |
                                REG_NOSUB) == 0)
        {
                if (regexec (&preg, str, 0, NULL, 0) == 0)
                        ret = TRUE;
                regfree (&preg);
        } else {
                DEBUG_PRINT ("There was an error compiling the perr_pattern regex: %s",perr_pattern);
        }
end:
        return ret;
}
/**
 * @desc reads file for pattern for known failure text messages
 * @param perr_pattern a valid regular expression pattern. @see man regex
 * @param filename a valid filename. Can be a special filename like "/dev/vcs1"
 * @return true if pattern was found
 */
inline gboolean
splashy_search_pattern (const gchar * perr_pattern, const gchar * filename, gshort use_ignore)
{
        FILE *fp;
        gchar *buffer = NULL;
        gint max_buf = 6200;    /* vcs is 6144 chars long. no new-lines are
                                 * added. man vcs */
        gboolean ret = FALSE;   /* assumes we will fail */

        if (!g_file_test (filename, G_FILE_TEST_EXISTS))
        {
                ERROR_PRINT ("Cannot test file %s\n", filename);
                return FALSE;
        }

        buffer = g_try_malloc (max_buf);
        if (buffer && (fp = g_fopen (filename, "r")))
        {
                fread_unlocked (buffer, max_buf, 1, fp);        /* *_unlocked 
                                                                 * are a GNU
                                                                 * extension: 
                                                                 * makes
                                                                 * function
                                                                 * fread()
                                                                 * thread
                                                                 * safe */
                fclose (fp);
                buffer[max_buf - 1] = '\0';
                ret = splashy_search_pattern (perr_pattern, buffer, use_ignore);
        }
        if (buffer)
                g_free (buffer);
        return ret;
}

/**
 * @desc gets current theme full path
 */
const gchar *
splashy_get_theme_dir ()
{
        GString *theme_path = g_string_new ("");
        const gchar *ret = NULL;
        /*
         * the current theme could be given to us in a full path or a 
         * simple string name. We determine how to get the theme
         * config file from this path 
         */
        const gchar *_current_theme = xml_parser_get_text (SPL_CURRENT_THEME);
        /*
         * is this a relative path ? 
         */
        if (g_ascii_strncasecmp (_current_theme, G_DIR_SEPARATOR_S, 1) != 0)
        {
                /*
                 * build path 
                 */
                g_string_append (theme_path,
                                 xml_parser_get_text (SPL_THEMES_DIR));
                g_string_append (theme_path, G_DIR_SEPARATOR_S);
                g_string_append (theme_path, _current_theme);
                /*
                 * FIXME g_free(_current_theme); 
                 */
        }
        else
        {
                /*
                 * use full path 
                 */
                g_string_append (theme_path, _current_theme);
        }
        ret = theme_path->str;
        /*
         * do not leak GString struct. free struct but not data
         * pointer/allocated mem 
         */
        g_string_free (theme_path, FALSE);
        return ret;
}

int open_nb(char *buf)
{
	int	fd, n;

	if ((fd = open(buf, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0)
		return -1;
	n = fcntl(fd, F_GETFL);
	n &= ~(O_NONBLOCK);
	fcntl(fd, F_SETFL, n);

	return fd;
}
