/*************************************************************************
 *                      splashy_config-functions.c                       *
 *                                                                       *
 *  Thu Aug 25 16:11:09 2005                                             *
 *  Copyright  2005  Vincent Amouret                                     *
 *  vincent.amouret@gmail.com                                            *
 *************************************************************************/
/*
 * This file is part of Splashy.
 * 
 * Splashy is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any
 * later version.
 * 
 * Splashy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with Splashy; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301 USA 
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>		/* system, exit */
#include <string.h>		/* strlen */
#include <unistd.h>		/* getuid */
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include <libgen.h>		/* basename */
#include <magic.h>		/* magic_* , mime type detection */

#include "common_macros.h"
#include "common_functions.h"
#include "xml_parser.h"
#include "xml_config.h"
#include "splashy_config-functions.h"

/*
 ** Private function
 */
void need_root (void);


/*
 ** set_new_theme
 ** arg: The name of the theme to apply
 ** desc: Change the theme used by Splashy
 */
void
set_new_theme (gchar * theme_name)
{
	need_root ();
	GString *new_theme =
		g_string_new (xml_parser_get_text (SPL_THEMES_DIR));
	g_string_append (new_theme, G_DIR_SEPARATOR_S);
	g_string_append (new_theme, theme_name);

	g_print (_(">Set theme as: "));
	g_print (theme_name);
	/*
	 ** Check if the theme exist
	 */
	if (g_file_test (new_theme->str, G_FILE_TEST_IS_DIR))
	{
		gen_config_xml ((gchar *)
				xml_parser_get_text (SPL_THEMES_DIR),
				theme_name,
				(gchar *)
				xml_parser_get_text (SPL_FIFO_XPATH),
				(gchar *)
				xml_parser_get_text ("/splashy/pid"));

		PRINT_DONE;
	}
	else
	{
		PRINT_FAIL;
	}

	g_string_free (new_theme, TRUE);
}

/*
 ** install_theme
 ** desc: install a theme from a tarball 
 ** arg: the tarball path 
 */
void
install_theme (gchar * tarball_path)
{
	need_root ();

	g_print (_(">Install theme"));

	if (g_file_test (tarball_path, G_FILE_TEST_EXISTS))
	{
		gboolean archive_is_valid = FALSE;
		const gchar *msg_invalid_format =
			_("This format is invalid\n");
		const gchar *mime_type;
		GString *command = g_string_new ("tar --extract ");

		magic_t cookie = magic_open (MAGIC_NONE);
		if (cookie == NULL)
		{
			printf ("ERROR: out of memory\n");
			return;
		}

		if (magic_load (cookie, NULL) == -1)
		{
			printf ("ERROR: %s\n", magic_error (cookie));
			return;
		}

		mime_type = magic_file (cookie, tarball_path);

		if (strncmp ("bzip", mime_type, 4) == 0)
		{
			g_string_append (command, " --bzip ");
		}
		else if (strncmp ("gzip", mime_type, 4) == 0)
		{
			g_string_append (command, " --gzip ");
		}
		else
		{
			if (g_strncasecmp (mime_type, "POSIX tar archive", 17)
			    != 0)
			{
				PRINT_FAIL;
				g_print (_
					 ("Did you really give me a tarball ?\n"));
				return;
			}
		}

		magic_close (cookie);

		g_string_append (command, " --file ");

		gchar *tmp_dir = g_build_filename (g_get_tmp_dir (),
						   "splashy-theme-XXXXXX",
						   NULL);
		tmp_dir = mkdtemp (g_strdup (tmp_dir));	// Create a temporary directory

		GString *tmp_command = g_string_new (command->str);
		g_string_append (tmp_command, tarball_path);
		g_string_append (tmp_command, " --directory ");
		g_string_append (tmp_command, tmp_dir);

		if (system (tmp_command->str) != 0)
		{
			PRINT_FAIL;
			g_print (_("An error occurred checking theme \n"));
		}
		else
		{
			// Get the theme name
			gchar *_theme = NULL;
			GError *ErrorDir = NULL;
			GDir *_dir = g_dir_open (tmp_dir, 0, &ErrorDir);
			if (ErrorDir != NULL)
			{
				g_assert (_dir == NULL);
				PRINT_FAIL;
				g_print ("%s", msg_invalid_format);
				fprintf (stderr, "%s\n", ErrorDir->message);
				g_error_free (ErrorDir);
				return;
			}
			_theme = g_strdup (g_dir_read_name (_dir));
			g_dir_close (_dir);

			// Remove the temporary directory                        
			const gchar *file = NULL;
			gchar *tmp_theme =
				g_build_filename (tmp_dir, _theme, NULL);

			// check if the archive contains SPL_THEME_CONFIG_FILE_NAME
			if (g_file_test
			    (g_build_filename
			     (tmp_theme, SPL_THEME_CONFIG_FILE_NAME, NULL),
			     G_FILE_TEST_IS_REGULAR))
				archive_is_valid = TRUE;

			_dir = g_dir_open (tmp_theme, 0, &ErrorDir);
			if (ErrorDir != NULL)
			{
				g_assert (_dir == NULL);
				PRINT_FAIL;
				g_print ("%s", msg_invalid_format);
				fprintf (stderr, "%s\n", ErrorDir->message);
				g_error_free (ErrorDir);
				return;
			}
			while ((file = g_dir_read_name (_dir)))
			{
				remove (g_strconcat
					(tmp_theme, G_DIR_SEPARATOR_S, file,
					 NULL));
			}
			rmdir (tmp_theme);
			rmdir (tmp_dir);
			g_dir_close (_dir);
			g_free (tmp_theme);
			g_free (tmp_dir);

			// Check if a theme with the same name exists
			gchar *target_path =
				g_build_filename (xml_parser_get_text
						  (SPL_THEMES_DIR), _theme,
						  NULL);

			if (g_strcasecmp
			    (target_path,
			     xml_parser_get_text (SPL_THEMES_DIR)) != 0)
			{
				if (g_file_test
				    (target_path, G_FILE_TEST_IS_DIR))
				{
					PRINT_FAIL;
					g_print (_
						 ("This theme already exists\n"));
					return;
				}
			}

			g_free (target_path);
			g_free (_theme);
		}

		if (archive_is_valid)
		{
			g_string_append (command, tarball_path);
			g_string_append (command, " --directory ");
			g_string_append (command,
					 xml_parser_get_text
					 (SPL_THEMES_DIR));

			// Really install the theme
			if (system (command->str) != 0)
			{
				PRINT_FAIL;
				g_print (_
					 ("An error occurred installing theme \n"));
			}
			else
			{
				PRINT_DONE;
			}
		}
		else
		{
			PRINT_FAIL;
			g_print (msg_invalid_format);
		}

		g_string_free (tmp_command, TRUE);
		g_string_free (command, TRUE);
	}
	else
	{
		PRINT_FAIL;
		g_print (_("The tarball doesn't exist !\n"));
	}
}

/*
 ** remove_theme
 ** desc: Remove a theme deleting his directory
 ** arg: the theme name
 */
void
remove_theme (gchar * theme_name)
{
	need_root ();
	g_print (_(">Remove theme"));
	GString *path = g_string_new (xml_parser_get_text (SPL_THEMES_DIR));
	g_string_append (path, G_DIR_SEPARATOR_S);
	g_string_append (path, theme_name);
	g_print (" %s", theme_name);


	const gchar *file = NULL;
	GString *buf = g_string_new ("");
	if (!g_file_test (path->str, G_FILE_TEST_IS_DIR))
	{
		PRINT_FAIL;
		g_print (_("The theme doesn't exist !\n"));
	}
	else if (g_strcasecmp
		 (theme_name, xml_parser_get_text (SPL_CURRENT_THEME)) == 0)
	{
		PRINT_FAIL;
		g_print (_("This theme is currently used !\n"));

	}
	else
	{
		DEBUG_PRINT ("splashy_config: Reading files from %s \n",
			     path->str);
		GError *ErrorDir = NULL;
		GDir *_dir = g_dir_open (path->str, 0, &ErrorDir);
		if (ErrorDir != NULL)
		{
			g_assert (_dir == NULL);
			PRINT_FAIL;
			fprintf (stderr, "%s\n", ErrorDir->message);
			g_error_free (ErrorDir);
			return;
		}
		while ((file = g_dir_read_name (_dir)))
		{

			g_string_assign (buf, path->str);
			g_string_append (buf, G_DIR_SEPARATOR_S);
			g_string_append (buf, file);
			DEBUG_PRINT ("splashy_config: Removing file %s \n",
				     buf->str);

			if (remove (buf->str) != 0)
			{
				PRINT_FAIL;
			}

		}
		g_dir_close (_dir);
		if (rmdir (path->str) != 0)
		{
			PRINT_FAIL;
		}
		else
		{
			PRINT_DONE;
		}
	}

	g_string_free (buf, TRUE);
	g_string_free (path, TRUE);
}

/*
 ** information
 ** desc: Display some informations
 */
void
information (void)
{
	g_print (_(">Theme currently used:\n"));
	g_print ("                      %s\n",
		 xml_parser_get_text ("/splashy/current_theme"));
	g_print ("                      version %s\n",
		 xml_parser_get_text ("/splashy/info/version"));
	g_print ("                         %s\n",
		 xml_parser_get_text ("/splashy/info/description"));
	g_print ("                      URLs %s\n",
		 xml_parser_get_text ("/splashy/info/urls"));
	g_print (_("                      by %s\n"),
		 xml_parser_get_text ("/splashy/info/author"));

	g_print (_(">Installed themes:\n"));

	const gchar *_theme = NULL;
	const gchar *themes_dir = xml_parser_get_text (SPL_THEMES_DIR);
	DEBUG_PRINT ("splashy_config: Reading themes from %s \n", themes_dir);
	GError *ErrorDir = NULL;
	GDir *_dir = g_dir_open (themes_dir, 0, &ErrorDir);
	if (ErrorDir != NULL)
	{
		g_assert (_dir == NULL);
		fprintf (stderr, "%s\n", ErrorDir->message);
		g_error_free (ErrorDir);
		return;
	}
	while ((_theme = g_dir_read_name (_dir)))
	{
		gchar *_theme_config = g_build_filename (themes_dir, _theme,
							 SPL_THEME_CONFIG_FILE_NAME,
							 NULL);
		DEBUG_PRINT ("splashy_config: testing theme dir %s\n",
			     _theme_config);
		if (g_file_test (_theme_config, G_FILE_TEST_IS_REGULAR))
		{
			g_print ("                      %s\n",
				 g_path_get_basename (_theme));
		}
	}
	g_dir_close (_dir);
}

/*
 ** A simple test if root priviledges are required
 */
void
need_root (void)
{
	if (getuid () != 0)
	{
		g_print (_
			 ("<!> Insufficient priviledges <!>\nRoot priviledges required for this function\n"));
		exit (0);
	}
}

/*
 ** create_theme
 ** desc: create a new theme in an interactive mode
 */
void
create_theme (XmlFields * NewTheme)
{
	g_print (_(">Create theme "));
	if (NewTheme->name != NULL)
	{
		gboolean font_file_exists = FALSE;
		const gchar *msg_file_doesnot_exists =
			_("This file doesn't exists: %s\n");
		gchar *theme_path =
			g_build_filename (xml_parser_get_text
					  (SPL_THEMES_DIR), NewTheme->name,
					  NULL);

		// Fields not used
		NewTheme->bg_x = NewTheme->bg_y = 0;

		// Check if a theme with the same name exists
		DEBUG_PRINT
			("splashy_config: Checking if '%s' already exists\n",
			 theme_path);
		if (g_strcasecmp
		    (theme_path, xml_parser_get_text (SPL_THEMES_DIR)) != 0)
		{
			if (g_file_test (theme_path, G_FILE_TEST_IS_DIR))
			{
				PRINT_FAIL;
				g_print (_("This theme already exists\n"));
				return;
			}
		}

		const gchar *mime_type;
		magic_t cookie = magic_open (MAGIC_MIME);
		if (cookie == NULL)
		{
			printf ("ERROR: out of memory\n");
			return;
		}

		if (magic_load (cookie, NULL) == -1)
		{
			printf ("ERROR: %s\n", magic_error (cookie));
			return;
		}


		// Check if bg_boot is an existing file
		DEBUG_PRINT
			("splashy_config: Checking if the boot image '%s' exists\n",
			 NewTheme->bg_boot);
		if (!g_file_test (NewTheme->bg_boot, G_FILE_TEST_EXISTS))
		{
			PRINT_FAIL;
			g_print (msg_file_doesnot_exists, NewTheme->bg_boot);
			return;
		}
		else
		{
			mime_type = magic_file (cookie, NewTheme->bg_boot);
			DEBUG_PRINT
				("splashy_config: Checking the boot image mime type: %s",
				 mime_type);
			if (g_strncasecmp (mime_type, "image/", 6) != 0)
			{
				PRINT_FAIL;
				g_print ("The boot image don't seem to be an image\n");
				return;
			}
		}

		// Check if bg_shutdown is an existing file
		DEBUG_PRINT
			("splashy_config: Checking if the shutdown image '%s' exists\n",
			 NewTheme->bg_shutdown);
		if (!g_file_test (NewTheme->bg_shutdown, G_FILE_TEST_EXISTS))
		{
			PRINT_FAIL;
			g_print (msg_file_doesnot_exists,
				 NewTheme->bg_shutdown);
			return;
		}
		else
		{
			mime_type =
				magic_file (cookie, NewTheme->bg_shutdown);
			DEBUG_PRINT
				("splashy_config: Checking the shutdown image mime type: %s",
				 mime_type);
			if (g_strncasecmp (mime_type, "image/", 6) != 0)
			{
				PRINT_FAIL;
				g_print ("The shutdown image don't seem to be an image\n");
				return;
			}
		}

		// Check if bg_error is an existing file
		DEBUG_PRINT
			("splashy_config: Checking if the error image '%s' exists\n",
			 NewTheme->bg_error);
		if (!g_file_test (NewTheme->bg_error, G_FILE_TEST_EXISTS))
		{
			PRINT_FAIL;
			g_print (msg_file_doesnot_exists, NewTheme->bg_error);
			return;
		}
		else
		{
			mime_type = magic_file (cookie, NewTheme->bg_error);
			DEBUG_PRINT
				("splashy_config: Checking the error image mime type: %s",
				 mime_type);
			if (g_strncasecmp (mime_type, "image/", 6) != 0)
			{
				PRINT_FAIL;
				g_print ("The error image don't seem to be an image\n");
				return;
			}
		}

		magic_close (cookie);	// no need magic anymore

		// Check the font file
		if (NewTheme->textfont_file == NULL)
		{
			NewTheme->textfont_file = "FreeSans.ttf";	// A dummy name
		}
		else
		{
			DEBUG_PRINT
				("splashy_config: Checking if '%s' exists (font file)\n",
				 NewTheme->textfont_file);
			if (!g_file_test
			    (NewTheme->textfont_file, G_FILE_TEST_EXISTS))
			{
				PRINT_FAIL;
				g_print (msg_file_doesnot_exists,
					 NewTheme->textfont_file);
				return;
			}
			else
			{
				font_file_exists = TRUE;
			}
		}
		//Default value below
		// Version number
		if (NewTheme->version <= 0)
		{
			NewTheme->version = 0.1;
		}

		// Description
		if (NewTheme->description == NULL)
		{
			NewTheme->description = g_strdup ("...");
		}

		// Url
		if (NewTheme->urls == NULL)
		{
			NewTheme->urls = g_strdup ("http://");
		}

		// Author
		if (NewTheme->author == NULL)
		{
			NewTheme->author =
				g_strdup ("Author <author@mail.com>");
		}

                // Background resolution: width
		if (NewTheme->bg_width < 0)
		{
			NewTheme->bg_width = 0;
		}

		// Background resolution: height
		if (NewTheme->bg_height < 0)
		{
			NewTheme->bg_height = 0;
		}
                
		// Progressbar: x coordinate
                if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
                {
                        if (NewTheme->pb_x < 0 || NewTheme->pb_x > 100)
                        {
                                NewTheme->pb_x = 0;
                        }
                }
                else if (NewTheme->pb_x < 0 )
                {
                        NewTheme->pb_x = 0;
                }

		// Progressbar: y coordinate
                if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
                {
                        if (NewTheme->pb_y < 0 || NewTheme->pb_y > 100)
                        {
                                NewTheme->pb_y = 0;
                        }
                }
                else if (NewTheme->pb_y < 0 )
                {
                        NewTheme->pb_y = 0;
                }

		// Progressbar: width
                if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
                {
                        if (NewTheme->pb_width < 0 || NewTheme->pb_width > 100)
                        {
                                NewTheme->pb_width = 0;
                        }
                }
                else if (NewTheme->pb_width < 0 )
                {
                        NewTheme->pb_width = 0;
                }

		// Progressbar: height
                if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
                {
                        if (NewTheme->pb_height < 0 || NewTheme->pb_height > 100)
                        {
                                NewTheme->pb_height = 0;
                        }
                }
                else if (NewTheme->pb_height < 0 )
                {
                        NewTheme->pb_height = 0;
                }

		// Progressbar: red color channel
		if (NewTheme->pb_red < 0 || NewTheme->pb_red > 255)
		{
			NewTheme->pb_red = 0;
		}

		// Progressbar: green color channel
		if (NewTheme->pb_green < 0 || NewTheme->pb_green > 255)
		{
			NewTheme->pb_green = 0;
		}

		// Progressbar: blue color channel
		if (NewTheme->pb_blue < 0 || NewTheme->pb_blue > 255)
		{
			NewTheme->pb_blue = 0;
		}

		// Progressbar: alpha color channel
		if (NewTheme->pb_alpha < 0 || NewTheme->pb_alpha > 255)
		{
			NewTheme->pb_alpha = 255;
		}

		// Progressbar: a border ?
		if (NewTheme->pb_border_show == NULL
		    || (g_strcasecmp (NewTheme->pb_border_show, "yes") != 0
			&& g_strcasecmp (NewTheme->pb_border_show,
					 "no") != 0))
		{
			NewTheme->pb_border_show = NULL;
			NewTheme->pb_border_show = g_strdup ("no");
		}

		// Progressbar: border red
		if (NewTheme->pb_border_red < 0
		    || NewTheme->pb_border_red > 255)
		{
			NewTheme->pb_border_red = 0;
		}

		// Progressbar: border green
		if (NewTheme->pb_border_green < 0
		    || NewTheme->pb_border_green > 255)
		{
			NewTheme->pb_border_green = 0;
		}

		// Progressbar: border blue
		if (NewTheme->pb_border_blue < 0
		    || NewTheme->pb_border_blue > 255)
		{
			NewTheme->pb_border_blue = 0;
		}

		// Progressbar: border alpha
		if (NewTheme->pb_border_alpha < 0
		    || NewTheme->pb_border_alpha > 255)
		{
			NewTheme->pb_border_alpha = 255;
		}

		

		// A textbox ?
		if (NewTheme->textbox_show == NULL
		    || (g_strcasecmp (NewTheme->textbox_show, "yes") != 0
			&& g_strcasecmp (NewTheme->textbox_show, "no") != 0))
		{
			NewTheme->textbox_show = g_strdup ("no");
		}

		// Textbox: x coordinate
		if (NewTheme->textbox_x < 0)
		{
			NewTheme->textbox_x = 0;
		}
		else if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
		{
			if (NewTheme->textbox_x > 100)
			{
				NewTheme->textbox_x = 0;
			}
		}

		// Textbox: y coordinate
		if (NewTheme->textbox_y < 0)
		{
			NewTheme->textbox_y = 0;
		}
		else if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
		{
			if (NewTheme->textbox_y > 100)
			{
				NewTheme->textbox_y = 0;
			}
		}

		// Textbox: width
		if (NewTheme->textbox_width < 0)
		{
			NewTheme->textbox_width = 0;
		}
		else if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
		{
			if (NewTheme->textbox_width > 100)
			{
				NewTheme->textbox_width = 0;
			}
		}

		// Textbox: height
		if (NewTheme->textbox_height < 0)
		{
			NewTheme->textbox_height = 0;
		}
		else if (NewTheme->bg_width <= 0 && NewTheme->bg_height <= 0)
		{
			if (NewTheme->textbox_height > 100)
			{
				NewTheme->textbox_height = 0;
			}
		}

		// Textbox: red color
		if (NewTheme->textbox_red < 0 || NewTheme->textbox_red > 255)
		{
			NewTheme->textbox_red = 0;
		}

		// Textbox: green color
		if (NewTheme->textbox_green < 0
		    || NewTheme->textbox_green > 255)
		{
			NewTheme->textbox_green = 0;
		}

		// Textbox: blue color
		if (NewTheme->textbox_blue < 0
		    || NewTheme->textbox_blue > 255)
		{
			NewTheme->textbox_blue = 0;
		}

		// Textbox: alpha channel
		if (NewTheme->textbox_alpha < 0
		    || NewTheme->textbox_alpha > 255)
		{
			NewTheme->textbox_alpha = 0;
		}

		// Textbox: a border ?
		if (NewTheme->textbox_border_show == NULL
		    || (g_strcasecmp (NewTheme->textbox_border_show, "yes") !=
			0
			&& g_strcasecmp (NewTheme->textbox_border_show,
					 "no") != 0))
		{
			NewTheme->textbox_border_show = g_strdup ("no");
		}

		// Textbox: border red
		if (NewTheme->textbox_border_red < 0
		    || NewTheme->textbox_border_red > 255)
		{
			NewTheme->textbox_border_red = 0;
		}

		// Textbox: border green
		if (NewTheme->textbox_border_green < 0
		    || NewTheme->textbox_border_green > 255)
		{
			NewTheme->textbox_border_green = 0;
		}

		// Textbox: border blue
		if (NewTheme->textbox_border_blue < 0
		    || NewTheme->textbox_border_blue > 255)
		{
			NewTheme->textbox_border_blue = 0;
		}

		// Textbox: border alpha
		if (NewTheme->textbox_border_alpha < 0
		    || NewTheme->textbox_border_alpha > 255)
		{
			NewTheme->textbox_border_alpha = 255;
		}

		// Font: height
		if (NewTheme->textfont_height < 0)
		{
			NewTheme->textfont_height = 10;
		}

		// Font: red
		if (NewTheme->textfont_red < 0
		    || NewTheme->textfont_red > 255)
		{
			NewTheme->textfont_red = 0;
		}

		// Font: green
		if (NewTheme->textfont_green < 0
		    || NewTheme->textfont_green > 255)
		{
			NewTheme->textfont_green = 0;
		}

		// Font: blue
		if (NewTheme->textfont_blue < 0
		    || NewTheme->textfont_blue > 255)
		{
			NewTheme->textfont_blue = 0;
		}

		// Font: alpha
		if (NewTheme->textfont_alpha < 0
		    || NewTheme->textfont_alpha > 255)
		{
			NewTheme->textfont_alpha = 0;
		}

		// Autoverbose on error ?
		if (NewTheme->verbose == NULL
		    || (g_strcasecmp (NewTheme->verbose, "yes") != 0
			&& g_strcasecmp (NewTheme->verbose, "no")))
		{
			NewTheme->verbose = g_strdup ("no");
		}

		// Progressbar on boot ?
		if (NewTheme->progressbaronboot == NULL
		    || (g_strcasecmp (NewTheme->progressbaronboot, "yes") != 0
			&& g_strcasecmp (NewTheme->progressbaronboot, "no")))
		{
			NewTheme->progressbaronboot = g_strdup ("yes");
		}

		// Progressbar on shutdown
		if (NewTheme->progressbaronshutdown == NULL
		    || (g_strcasecmp (NewTheme->progressbaronshutdown, "yes")
			!= 0
			&& g_strcasecmp (NewTheme->progressbaronshutdown,
					 "no")))
		{
			NewTheme->progressbaronshutdown = g_strdup ("yes");
		}

		// Fade in ?
		if (NewTheme->fadein == NULL
		    || (g_strcasecmp (NewTheme->fadein, "yes")
			!= 0 && g_strcasecmp (NewTheme->fadein, "no")))
		{
			NewTheme->fadein = g_strdup ("yes");
		}

		// Fade out ?
		if (NewTheme->fadeout == NULL
		    || (g_strcasecmp (NewTheme->fadeout, "yes")
			!= 0 && g_strcasecmp (NewTheme->fadeout, "no")))
		{
			NewTheme->fadeout = g_strdup ("yes");
		}

		// Create the target directory
		if (g_mkdir (theme_path, 0755) != 0)
		{
			g_print (_("Error! Can't create directory %s\n"),
				 theme_path);
			g_print (_("Writing theme %s to current directory\n"),
				 NewTheme->name);
			if (g_mkdir (NewTheme->name, 0755) != 0)
			{
				g_print (_
					 ("Error! Can't create directory %s\n"),
					 NewTheme->name);
				exit (1);
			}
			else
			{
				theme_path = g_strdup (NewTheme->name);
			}
		}

		/*
		 * upload the boot image 
		 */
		gchar *dest = g_build_filename (theme_path,
						g_basename (NewTheme->
							    bg_boot),
						NULL);
		if (!copy_file (NewTheme->bg_boot, dest))
		{
			g_print (_
				 ("Error! Unable to use the picture %s.\nAborting\n"),
				 dest);
			remove_theme (NewTheme->name);
			return;
		}
		else if (g_path_is_absolute (dest))
		{
			NewTheme->bg_boot = g_strdup (dest);
		}
		else
		{
			NewTheme->bg_boot =
				g_strdup (g_basename (NewTheme->bg_boot));
		}

		/*
		 * upload the shutdown image 
		 */
		dest = g_build_filename (theme_path,
					 g_basename (NewTheme->bg_shutdown),
					 NULL);
		if (!copy_file (NewTheme->bg_shutdown, dest))
		{
			g_print (_
				 ("Error! Unable to use the picture %s.\nAborting\n"),
				 dest);
			remove_theme (NewTheme->name);
			return;
		}
		else if (g_path_is_absolute (dest))
		{
			NewTheme->bg_shutdown = g_strdup (dest);
		}
		else
		{
			NewTheme->bg_shutdown =
				g_strdup (g_basename (NewTheme->bg_shutdown));
		}

		/*
		 * upload the error image 
		 */
		dest = g_build_filename (theme_path,
					 g_basename (NewTheme->bg_error),
					 NULL);
		if (!copy_file (NewTheme->bg_error, dest))
		{
			g_print (_
				 ("Error! Unable to use the picture %s.\nAborting\n"),
				 dest);
			remove_theme (NewTheme->name);
			return;
		}
		else if (g_path_is_absolute (dest))
		{
			NewTheme->bg_error = g_strdup (dest);
		}
		else
		{
			NewTheme->bg_error =
				g_strdup (g_basename (NewTheme->bg_error));
		}

		/*
		 * upload the font file
		 */
		if (font_file_exists)
		{
			dest = g_build_filename (theme_path,
						 g_basename (NewTheme->
							     textfont_file),
						 NULL);
			if (!copy_file (NewTheme->textfont_file, dest))
			{
				g_print (_
					 ("Error! Unable to use the picture %s.\nAborting\n"),
					 dest);
				remove_theme (NewTheme->name);
				return;
			}
			else
			{
				NewTheme->textfont_file =
					g_strdup (g_basename
						  (NewTheme->textfont_file));
			}
		}

		gchar *theme_file_path = g_build_filename (theme_path,
							   SPL_THEME_CONFIG_FILE_NAME,
							   NULL);

		g_print (NewTheme->name);


		if (gen_theme_xml (theme_file_path, NewTheme) == 0)
			PRINT_DONE;
		else
			PRINT_FAIL;

		g_free (theme_file_path);
		g_free (dest);
		g_free (theme_path);
	}
	else
	{
		PRINT_FAIL;
		g_print (_("You try to create a theme without a name\n"));
		return;
	}
}

/*
 ** get_fields 
 ** desc: The interactive mode for create_theme
 */
XmlFields *
get_fields (void)
{
	XmlFields *NewTheme = g_new0 (XmlFields, 1);
	/*    size_t n = 0;
	 * 
	 * g_print (_("--    Theme    --\n"));
	 * 
	 * g_print (_(">Name       : "));
	 * scanf ("%as", &NewTheme->name);
	 * 
	 * g_print (_(">Version       : "));
	 * scanf ("%f", &NewTheme->version);
	 * getchar ();             // For the \n after the scanf()
	 * getchar ();             // For the \n after the scanf()
	 * getchar ();             // For the \n after the scanf()
	 * 
	 * g_print (_(">Description: "));
	 * n = 0;
	 * getline (&NewTheme->description, &n, stdin);
	 * NewTheme->description[strlen (NewTheme->description) - 1] = '\0';
	 * 
	 * g_print (_(">URLs: "));
	 * n = 0;
	 * getline (&NewTheme->urls, &n, stdin);
	 * NewTheme->urls[strlen (NewTheme->urls) - 1] = '\0';
	 * 
	 * g_print (_(">Author     : "));
	 * n = 0;
	 * getline (&NewTheme->author, &n, stdin);
	 * NewTheme->author[strlen (NewTheme->author) - 1] = '\0';
	 * 
	 * g_print (_("-- ProgressBar --\n"));
	 * g_print (_("**  Position   **\n"));
	 * g_print ("**   0->100    **\n");
	 * 
	 * g_print (_(">X          : "));
	 * scanf ("%u", &NewTheme->x);
	 * 
	 * g_print (_(">Y          : "));
	 * scanf ("%u", &NewTheme->y);
	 * 
	 * g_print (_(">Width      : "));
	 * scanf ("%u", &NewTheme->width);
	 * 
	 * g_print (_(">Height     : "));
	 * scanf ("%u", &NewTheme->height);
	 * 
	 * g_print (_("**    Color    **\n"));
	 * g_print ("**    0->255   **\n");
	 * 
	 * g_print (_(">Red        : "));
	 * scanf ("%u", &NewTheme->red);
	 * 
	 * g_print (_(">Green      : "));
	 * scanf ("%u", &NewTheme->green);
	 * 
	 * g_print (_(">Blue       : "));
	 * scanf ("%u", &NewTheme->blue);
	 * 
	 * g_print (_(">Alpha      : "));
	 * scanf ("%u", &NewTheme->alpha);
	 * 
	 * g_print (_(">Border"));
	 * g_print (" [yes|no]: ");
	 * scanf ("%as", &NewTheme->border_show);
	 * g_print (_(">Border Red"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->border_red);
	 * g_print (_(">Border Green"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->border_green);
	 * g_print (_(">Border Blue"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->border_blue);
	 * g_print (_(">Border Alpha"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->border_alpha);
	 * 
	 * g_print (_("-- Background --\n"));
	 * g_print (_("** Path to    **\n"));
	 * g_print (_("**upload files**\n"));
	 * g_print (_(">Boot       : "));
	 * scanf ("%as", &NewTheme->boot);
	 * 
	 * g_print (_(">Shutdown   : "));
	 * scanf ("%as", &NewTheme->shutdown);
	 * 
	 * g_print (_(">Error      : "));
	 * scanf ("%as", &NewTheme->error);
	 * 
	 * g_print (_("-- Text Area --\n"));
	 * g_print (_("--   Position  --\n"));
	 * g_print (_(">X      : "));
	 * scanf ("%u", &NewTheme->textarea_x);
	 * 
	 * g_print (_(">Y      : "));
	 * scanf ("%u", &NewTheme->textarea_y);
	 * 
	 * g_print (_(">Width      : "));
	 * scanf ("%u", &NewTheme->textarea_width);
	 * 
	 * g_print (_(">Height      : "));
	 * scanf ("%u", &NewTheme->textarea_height);
	 * 
	 * g_print (_("--   Color  --\n"));
	 * g_print (_(">Red      : "));
	 * scanf ("%u", &NewTheme->textarea_red);
	 * g_print (_(">Green      : "));
	 * scanf ("%u", &NewTheme->textarea_green);
	 * g_print (_(">Blue      : "));
	 * scanf ("%u", &NewTheme->textarea_blue);
	 * g_print (_(">Alpha        : "));
	 * scanf ("%u", &NewTheme->textarea_alpha);
	 * 
	 * g_print (_(">Border"));
	 * g_print (" [yes|no]: ");
	 * scanf ("%as", &NewTheme->textarea_border_show);
	 * g_print (_(">Border Red"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->textarea_border_red);
	 * g_print (_(">Border Green"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->textarea_border_green);
	 * g_print (_(">Border Blue"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->textarea_border_blue);
	 * g_print (_(">Border Alpha"));
	 * g_print (" [0-255]: ");
	 * scanf ("%u", &NewTheme->textarea_border_alpha);
	 * 
	 * g_print (_("-- Text Font--\n"));
	 * 
	 * g_print (">Filename : ");
	 * scanf ("%as", &NewTheme->textfont_file);
	 * 
	 * g_print (_(">Height      : "));
	 * scanf ("%u", &NewTheme->textfont_height);
	 * 
	 * g_print (_(">Red      : "));
	 * scanf ("%u", &NewTheme->textfont_red);
	 * 
	 * g_print (_(">Green      : "));
	 * scanf ("%u", &NewTheme->textfont_green);
	 * 
	 * g_print (_(">Blue      : "));
	 * scanf ("%u", &NewTheme->textfont_blue);
	 * 
	 * g_print (_(">Alpha        : "));
	 * scanf ("%u", &NewTheme->textfont_alpha);
	 * 
	 * g_print (_(">Autoverbose"));
	 * g_print (" [yes | no]: ");
	 * scanf ("%as", &NewTheme->verbose);
	 * 
	 * g_print (_(">Progressbar on boot"));
	 * g_print (" [yes | no]: ");
	 * scanf ("%as", &NewTheme->progressbaronboot);
	 * 
	 * g_print (_(">Progressbar on shutdown"));
	 * g_print (" [yes | no]: ");
	 * scanf ("%as", &NewTheme->progressbaronshutdown);
	 * 
	 * if (g_strcasecmp (NewTheme->border_show, "yes") != 0
	 * && g_strcasecmp (NewTheme->border_show, "no") != 0)
	 * g_stpcpy (NewTheme->border_show, "no");
	 * 
	 * if (g_strcasecmp (NewTheme->verbose, "yes") != 0
	 * && g_strcasecmp (NewTheme->verbose, "no") != 0)
	 * g_stpcpy (NewTheme->verbose, "yes");
	 * 
	 * if (g_strcasecmp (NewTheme->progressbaronboot, "yes") != 0
	 * && g_strcasecmp (NewTheme->progressbaronboot, "no") != 0)
	 * g_stpcpy (NewTheme->progressbaronboot, "yes");
	 * 
	 * if (g_strcasecmp (NewTheme->progressbaronshutdown, "yes") != 0
	 * && g_strcasecmp (NewTheme->progressbaronshutdown, "no") != 0)
	 * g_stpcpy (NewTheme->progressbaronshutdown, "yes");
	 */
	return NewTheme;
}

/*
 ** upload_image
 ** desc: upload a file
 ** args: src_path the source file
 **       dest_path the destination file
 */
gboolean
copy_file (gchar * src_path, gchar * dest_path)
{
	FILE *src_file;
	FILE *dest_file;
	struct stat file;
	gulong filesize = 0;
	gchar *content;


	if ((src_file = fopen (src_path, "r")) == NULL)
	{
		// g_critical("Error opening file %s\n", src_path);
		return FALSE;
	}


	if (!g_stat (src_path, &file))
	{
		filesize = file.st_size;
	}
	content = (gchar *) g_malloc (filesize);
	fread (content, 1, filesize, src_file);
	fclose (src_file);

	if ((dest_file = fopen (dest_path, "w+")) == NULL)
	{
		g_critical ("Error opening dest file %s\n", dest_path);
		return FALSE;
	}
	fwrite (content, 1, filesize, dest_file);
	fclose (dest_file);

	g_free (content);
	return TRUE;
}
