/************************************************************************
 *                              functions.c                             *
 *                                                                      *
 *  2005-05-28 03:01 EDT                                                *
 *  Copyright  2005  Otavio Salvador <otavio@debian.org>                *
 *                   Luis Mondesi    <lemsx1@gmail.com>                 *
 ************************************************************************/
/*
 * This file is part of Splashy.
 * 
 * Splashy is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any
 * later version.
 * 
 * Splashy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with Splashy; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301 USA 
 *
 * Bugs: 
 *  - openpty() is only in glibc
 *  - uses VT_* calls on Linux
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*
 * Standard headers 
 */
#ifndef _GNU_SOURCE
/*
 * Allow GNU specific funtions: *_unlocked() in stdio.h 
 */
#define _GNU_SOURCE
#endif /* _GNU_SOURCE */
#include <errno.h>              /* needed by all file manipulating functions */
#include <stdio.h>              /* NULL and others */
#include <stdlib.h>             /* strtol() exit() */
#include <unistd.h>             /* fork() getpid() unlink() */
#include <pthread.h>            /* pthread functions and data structures */
/*
 * TODO get rid of string.h. See FIXME's 
 */
#include <string.h>             /* strlen() and str*() fam */
#include <signal.h>             /* kill() handle SIGHUP SIGUSR1 SIGUSR2 */
#include <time.h>               /* nanosleep() */
#include <glib.h>               /* too many... */
#include <glib/gstdio.h>        /* g_fopen() */
#include <fcntl.h>              /* open() */
#include <sys/ioctl.h>          /* ioctl() */
#ifdef __linux__
/*
 * #include <sys/mount.h> USE for mounting /proc if not mounted yet 
 */
#include <linux/vt.h>           /* VT_* */
#endif

#include "common_macros.h"
#include "common_functions.h"
#include "xml_parser.h"
#include "xml_config.h"
#include "splashy_functions.h"
#include "splashy_video.h"

static gint last_progress = 0;  /* how many exact pixels to draw in the
                                 * progressbar (0-100). Lock before using
                                 * this value. */

static gint arg_progress = 0;   /* value to be sent to last_progress from
                                 * FIFO. @see keyevent_loop() */

static gboolean _show_progress = FALSE;
static splashy_video_t video;

static gboolean exiting = FALSE;        /* threads should read this before
                                         * doing anything at all. this is set 
                                         * to true at cmd_exit() */
static gboolean read_console = FALSE;   /* flag to turn on/off reading the
                                         * /dev/vcs* files */
static gboolean show_textbox_area = FALSE;      /* flag to toggle showing
                                                 * textbox area */
static gboolean switched_to_verbose = FALSE;    /* flag to know when verbose
                                                 * image is displayed */
static gboolean F2_toggle_pressed = FALSE;      /* keeps track of the depress 
                                                 * F2 key */
static const gchar *_current_background = NULL; /* when pressing F2 we want
                                                 * to go back to our original 
                                                 * * background */
static gint pipe_fd;            /* our named pipe splashy.fifo */
static gint timeout = 120;      /* time out when reading from splashy.fifo
                                 * (in seconds). If splashy-fifo is not
                                 * written to for this long, splashy will
                                 * exit */

/*
 * functions 
 */
gint
cmd_allow_chvt (void **args)
{
        DEBUG_PRINT ("%s\n", "cmd_allow_chvt() called");
        video_allow_vt_switching ();
        return 0;
}

gint
cmd_exit (void **args)
{
        /*
         * TODO: getting splashy to switch consoles (tty's) automatically
         * after finishing booting turned out to be a bad idea. I'm living
         * the code here commented out just in case we find a solution for
         * this issue later. For now, we wait until we exit, and then send a
         * chvt from splashy99 when splashy has already exited. Users should
         * customize the tty to which splashy will be switched from
         * /etc/default/splashy (on debian). Nonetheless, we let
         * video_allow_vt_switching() here just don't set vt-switching
         * directfb option there as this breaks splashy. - Luis Mondesi
         * <lemsx1@gmail.com> 
         */
        DEBUG_PRINT ("cmd_exit called %s", "");
        if ( exiting == TRUE ) 
                return -1; /* do not allow this function to be called more than once */
        exiting = TRUE;
        DEBUG_PRINT ("cmd_exit() exiting set to TRUE %s", "");

        /* 
         * TODO
         * video_allow_vt_switching ();
         */
        read_console = FALSE;   /* stop updating textbox area */

        /*
         * do an expectacular exit routine 
         */
        const gchar *fadeout = xml_parser_get_text ("/splashy/fadeout");
        if (fadeout)
        {
                if (g_ascii_strncasecmp (fadeout, "yes", 3) == 0)
                {
                        DEBUG_PRINT ("cmd_exit fadeout routine %s", "");
                        video_fade_out (&video);
                }
        }

        DEBUG_PRINT ("cmd_exit removing splashy temp files and pid %s", "");

        const gchar *_pid_file = xml_parser_get_text ("/splashy/pid");
        if (g_file_test (_pid_file, G_FILE_TEST_EXISTS))
        {
                unlink (_pid_file);
        }

        /*
         * No need to remove the fifo, the reasoning is that this file can
         * be used from read-only partitions only if it exist. You can't
         * create it. That means that users should probably always run "splashy test"
         * right after the first installation. Splashy will then create the FIFO
         * in the right path. - Luis Mondesi
         
         const gchar *_fifo_file = xml_parser_get_text (SPL_FIFO_XPATH);
         if (g_file_test (_fifo_file, G_FILE_TEST_EXISTS))
         {
         unlink (_fifo_file);
         }*/

        DEBUG_PRINT ("cmd_exit releasing memory %s", "");

        /*
         * free up memory 
         */
        video.ev_buffer->Release (video.ev_buffer);     /* input buffer */
        video.keyboard->Release (video.keyboard);       /* keyevents gone */
        video.primary_surface->Release (video.primary_surface); /* pix holder 
                                                                 */
        video.primary_window->Release (video.primary_window);
        video.dfb->Release (video.dfb); /* kill it ! */

        exit (0);
        return 0;               /* we never reach this */
}

gint
cmd_progress (void **args)
{
        if (*(gint *) args[0] < 0 || *(gint *) args[0] > PROGRESS_MAX)
                return -1;

        arg_progress = *(gint *) args[0];

        if (arg_progress <= 0)
                return 1;
        /*
         * allow time for display to be drawn 
         */

        if (last_progress != arg_progress)
        {
                DEBUG_PRINT ("cmd_progress: Setting progressbar to %d ticks",
                             arg_progress);
                video_update_progressbar (&video, &video.progressbar,
                                          arg_progress);
        }

        last_progress = arg_progress;

        sched_yield ();

        return 0;
}

gint
cmd_print (void **args)
{
        video_printline (&video, args[0], _show_progress);
        return 0;
}

gint
cmd_print_clear (void **args)
{
        video_printline (&video, "", _show_progress);
        return 0;
}

gint
cmd_chvt (void **args)
{
        /*
         * vt[1-7] 
         */
        if (*(gint *) args[0] < 1 || *(gint *) args[0] > 7)
                return -1;
        video_allow_vt_switching ();    /* tell libdirectfb that's ok to
                                         * allow vt switching */
        splashy_chvt (*(gint *) args[0]);
        return 0;
}

gint
cmd_set_timeout (void **args)
{
        if (*(gint *) args[0] < 1)
                return -1;

        timeout = *(gint *) args[0];
        sched_yield ();

        return 0;
}

gint
cmd_repaint (void **args)
{
        /*
         * TODO change background to current image 
         */
        /*
         * video_printline (&video, "", _show_progress); 
         */
        return 0;
}


/**
 * @desc our command-holding object and a list of supported commands
 *
 * TODO implement the rest of the commands
 */
struct
{
        const gchar *cmd;
          gint (*handler) (void **);
        gint args;
        gchar *specs;
} known_cmds[] =
{
        {
        .cmd = "timeout",.handler = cmd_set_timeout,.args = 1,.specs =
                        "d"},
        {
        .cmd = "repaint",.handler = cmd_repaint,.args = 0,.specs =
                        NULL,},
        {
        .cmd = "progress",.handler = cmd_progress,.args = 1,.specs =
                        "d"},
        {
        .cmd = "PROGRESS",.handler = cmd_progress,.args = 1,.specs =
                        "d"},
        {
        .cmd = "print",.handler = cmd_print,.args = 1,.specs = "s"},
        {
        .cmd = "TEXT",.handler = cmd_print,.args = 1,.specs = "s"},
        {
        .cmd = "CLEAR",.handler = cmd_print_clear,.args =
                        0,.specs = NULL},
        {
        .cmd = "clrprint",.handler = cmd_print_clear,.args =
                        0,.specs = NULL},
        {
        .cmd = "chvt",.handler = cmd_chvt,.args = 1,.specs = "d"},
        {
        .cmd = "allowchvt",.handler = cmd_allow_chvt,.args =
                        0,.specs = NULL},
        {
        .cmd = "exit",.handler = cmd_exit,.args = 0,.specs = NULL},
        {
        .cmd = "QUIT",.handler = cmd_exit,.args = 0,.specs = NULL}
};

/*
 * helper for signal handling 
 */
void
sig_exit_safe (gint sig)
{
        g_printerr ("Splashy caught signal number %d. Exiting...", sig);
        cmd_exit (NULL);
}

void
sig_hup_handler (gint sig)
{
        int i = 0;
        g_printerr
                ("Splashy caught HUP signal number %d. Closing open FD's...",
                 sig);
        /*
         * close all open filehandles
         */
        for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
        {
                if (i == 2)
                        continue;       /* never close stderr */
                DEBUG_PRINT ("sig_hup_handler: Filehandle closed %d", i);
                close (i);
        }
}

/**
 * @desc a simple function that waits for text to be written to a named pipe and dispatches commands accordingly
 * @param _fifo named pipe to read
 * @return
 */
inline void
spl_read_fifo (FILE * _fifo)
{
        DEBUG_PRINT ("Reading fifo %s...\n", "");
        gchar buf[1024]; /** used as buffer for reading FIFO */
        gint i, j, k;

        /*
         * Read FIFO and call functor to handle exceptions 
         */
        while (fgets_unlocked (buf, 1024, _fifo))
        {
                gchar *t;
                int args_i[4];
                void *args[4];

                /*
                 * buf[1023] = 0; 
                 */
                /*
                 * FIXME buf[g_utf8_strlen(buf,1024)-1] = 0; 
                 */
                /*
                 * buf[strlen (buf) - 1] = 0; 
                 */

                for (i = 0; i < sizeof (known_cmds) / sizeof (known_cmds[0]);
                     i++)
                {
                        k = strlen (known_cmds[i].cmd);
                        /*
                         * FIXME k = g_utf8_strlen(known_cmds[i].cmd,1024); 
                         */

                        /*
                         * FIXME if (g_ascii_strncasecmp(buf,
                         * known_cmds[i].cmd, k)) 
                         */
                        if (strncmp (buf, known_cmds[i].cmd, k))
                                continue;

                        for (j = 0; j < known_cmds[i].args; j++)
                        {

                                for (; buf[k] == ' '; buf[k] = 0, k++);
                                if (!buf[k])
                                        goto out;

                                switch (known_cmds[i].specs[j])
                                {

                                case 's':
                                        args[j] = &(buf[k]);
                                        for (; buf[k] != ' '; k++);
                                        break;

                                case 'd':
                                        /*
                                         * FIXME args_i[j] =
                                         * g_ascii_strtod(&(buf[k]), &t); 
                                         */
                                        args_i[j] =
                                                strtol (&(buf[k]), &t, 10);
                                        if (t == &(buf[k]))
                                                goto out;

                                        args[j] = &(args_i[j]);
                                        k = t - buf;
                                        break;
                                }
                        }
                        /**
                         * cmds should be used for everything
                         * @see known_cmds
                         * and handlers
                         */
                        known_cmds[i].handler (args);
                }
        }
      out:fclose (_fifo);
}

/**
 * @desc a helper function to construct a filename path from a given xpath
 * It reads config.xml for a given theme and builds the path accordingly
 * @param theme_xpath use this theme
 * @param image_xpath path representing string to retrieve from xml file
 * @return full path to image
 */
static inline const gchar *
_splashy_theme_image_path (const gchar * theme_xpath,
                           const gchar * image_xpath)
{
        DEBUG_PRINT ("_splashy_theme_image_path(theme_xpath,%s) called",
                     image_xpath);
        GString *_ret_path = g_string_new ("");

        const gchar *_path = xml_parser_get_text (image_xpath);

        /*
         * is this a relative path? 
         */
        if (g_ascii_strncasecmp (_path, "/", 1) != 0)
        {
                /*
                 * root dir to themes: i.e. /etc/splashy/themes 
                 */
                g_string_append (_ret_path,
                                 xml_parser_get_text (SPL_THEMES_DIR));
                g_string_append (_ret_path, "/");
                /*
                 * current theme name 
                 */
                g_string_append (_ret_path,
                                 xml_parser_get_text (theme_xpath));
                g_string_append (_ret_path, "/");
                if (!g_file_test (_ret_path->str, G_FILE_TEST_IS_DIR))
                {
                        ERROR_PRINT ("Directory '%s' does not exist",
                                     _ret_path->str);
                        return NULL;
                }
                /*
                 * image name 
                 */
                g_string_append (_ret_path, _path);
        }
        else
        {
                /*
                 * assume this is the full path to the image 
                 */
                g_string_append (_ret_path, _path);
        }
        return _ret_path->str;
}

inline void
_switch_to_verbose_image ()
{
        const gchar *background =
                _splashy_theme_image_path (SPL_CURRENT_THEME,
                                           "/splashy/background/errorimg");
        if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
                background =
                        _splashy_theme_image_path
                        (SPL_DEFAULT_THEME, "/splashy/background/errorimg");
        video_change_splash (&video, background, _show_progress);

        /*
         * safe for future reference 
         */
        _current_background = background;
}

/**
 * When pressing F2, or when autoverboseonerror is set
 * we show the scrolling UNIX text from the console /dev/vcs1, in a smaller window
 * if the user needs to type something, s/he must press ESC to exit Splashy
 * @param data - not used
 * @return
 *
 * TODO
 * - make use of real console
 * - this is a slow function
 */
inline void *
verbose_text_loop (void *data)
{
        FILE *dev_vcs;
        char buf[81];
        GString *buf_str;       /* what to display */
        GString *device;        /* /dev/vcs to read */
        gshort i, j;
        const gchar *error_pattern;

        gboolean found_error;
        gboolean autoverbose;
        struct timespec _sleep;

        dev_vcs = NULL;
        switched_to_verbose = FALSE;    /* verbose image flag */
        show_textbox_area = FALSE;      /* when this flag is set we display a 
                                         * textbox */
        autoverbose = FALSE;    /* assume we don't need verbose mode for now */
        read_console = TRUE;    /* assume we will be able to read from a
                                 * console */
        found_error = FALSE;    /* we have not found any errors yet */
        error_pattern = "(FATAL|fail|error|abort|===|recovering journal)";

        _sleep.tv_sec = 1;
        _sleep.tv_nsec = 0;

        device = g_string_new ("");

        /*
         * get autoverboseonerror from config file and set autoverbose
         * accordingly 
         */
        if (g_ascii_strncasecmp
            (xml_parser_get_text (SPL_AUTO_VERBOSE_ON_ERROR), "yes", 3) == 0)
                autoverbose = TRUE;

        /*
         * set the cancellation parameters -- - Enable thread cancellation - 
         * Defer the action of the cancellation 
         */
        pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
        pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, NULL);

        while (1)
        {
                if (exiting || read_console == FALSE)
                {
                        DEBUG_PRINT ("Not reading console at all...%s", "");
                        /*
                         * this thread shouldn't be holding a lock on this
                         * but... 
                         */
                        sched_yield ();
                        nanosleep (&_sleep, NULL);
                        continue;
                }
                /*
                 * read the realcons console and look for error patterns 
                 * on error, set our background image accordingly and display
                 * text printing 1 line (80 char) at a time
                 */
                /*
                 * Ubuntu uses /dev/vcs2, every other distro /dev/vcs1 
                 */
                for (i = 1; i <= 7; i++)
                {
                        g_string_printf (device, "/dev/vcs%d", i);
                        if (!g_file_test (device->str, G_FILE_TEST_EXISTS))
                        {
                                sched_yield ();
                                continue;
                        }
                        dev_vcs = g_fopen (device->str, "r");
                        if (!dev_vcs)
                        {
                                DEBUG_PRINT ("Can't open %s for reading",
                                             device->str);
                                sched_yield ();
                                /*
                                 * short wait 
                                 */
                                _sleep.tv_sec = 0;
                                /*
                                 * 1/2 second 
                                 */
                                _sleep.tv_nsec = 459999999;
                                nanosleep (&_sleep, NULL);
                                continue;
                        }
                        while (fgets_unlocked (buf, 81, dev_vcs))
                        {
                                if (!buf || strlen (buf) < 8)
                                        continue;
                                /*
                                 * we need to find where the spaces end 
                                 */
                                for (j = 0; j < strlen (buf); j++)
                                {
                                        if (buf[j] != ' ')
                                        {
                                                break;
                                        }
                                }

                                /*
                                 * copy only the characters now (no leading spaces)
                                 */

                                buf_str = g_string_new ("");
                                for (; j < strlen (buf); j++)
                                {
                                        g_string_append_c (buf_str, buf[j]);
                                }

                                /*
                                 * we don't need to print blank lines or new-lines
                                 */
                                /*
                                 * if (strlen(buf) < 8)
                                 */
                                if (splashy_search_pattern_str
                                                ("[^a-zA-Z0-9]", buf_str->str,
                                                 0) == FALSE)
                                {

                                        g_string_free (buf_str, TRUE);
                                        continue;
                                }

                                /*
                                 * look for error messages only if we weren't asked
                                 * to display text right the way
                                 */
                                if (show_textbox_area != TRUE)
                                {
                                        /*
                                         * we only look for errors if autoverbose
                                         * is set to TRUE
                                         * once an error is found, we keep printing
                                         * the text to the textbox area without
                                         * matching any more errors
                                         */

                                        if (found_error == FALSE)
                                        {
                                                if (autoverbose == TRUE)
                                                {
                                                        found_error =
                                                                splashy_search_pattern_str
                                                                (error_pattern,
                                                                 buf, 1);
                                                }

                                                /* when the user press F2 we want to show
                                                 * the buffer in the textbox anyway. regardless
                                                 * of whether an error was found or not */
                                                if (F2_toggle_pressed != TRUE)
                                                {
                                                        g_string_free (buf_str, TRUE);
                                                        continue;
                                                }
                                        }
                                        else
                                        {
                                                /*
                                                 * error was found, we need to display the error image
                                                 * and allow the textbox area to be shown
                                                 */
                                                found_error = TRUE;
                                                show_textbox_area = TRUE;

                                                if (autoverbose == TRUE
                                                    && switched_to_verbose ==
                                                    FALSE)
                                                {
                                                        _switch_to_verbose_image
                                                                ();
                                                        switched_to_verbose =
                                                                TRUE;
                                                        /*
                                                         * fake F2 pressed so that users can press
                                                         * this key to turn off the textbox area
                                                         */
                                                        F2_toggle_pressed =
                                                                TRUE;
                                                }
                                        }
                                }

                                /*
                                 * pressing F2 will cause text to be printed to the textbox
                                 * as well as having autoverbose set to on in the config.xml
                                 */
                                if (autoverbose == TRUE
                                    || F2_toggle_pressed == TRUE)
                                {
                                        if (!exiting)
                                                video_printline_s (&video,
                                                                   (char *)
                                                                   buf_str->
                                                                   str,
                                                                   _show_progress);

                                        sched_yield ();

                                        for (j = 0; j < 81; j++)
                                        {
                                                buf[j] = '\0';  /* make it
                                                                 * all NULs */
                                        }

                                        /*
                                         * wait for 1/6 seconds
                                         */
                                        _sleep.tv_sec = 0;
                                        _sleep.tv_nsec = 114999999;
                                        nanosleep (&_sleep, NULL);
                                }
                                g_string_free (buf_str, TRUE);
                        }
                }
        }
        g_string_free (device, TRUE);   /* never reached */
        pthread_exit (NULL);    /* never reached */
}
inline void *
fifo_loop (void *data)
{
        gboolean _preview = *(gboolean *) data;
        int err, i = 0;
        fd_set descriptors;
        struct timeval tv;
        struct timespec _sleep;

        gboolean forward_flag = TRUE;   /* this is only used when testing. we 
                                         * always start forward when testing */

        /*
         * FIXME video_set_progressbar_forward(forward_flag);
         */

        /*
         * It's important that the progress bar code runs as a child process
         * because this doesn't need to grab events from the keyboard. All
         * communication to this is done through a FIFO 
         */

        const gchar *_spl_fifo_path = xml_parser_get_text (SPL_FIFO_XPATH);
        if (!_spl_fifo_path)
        {
                /*
                 * does it make any sense at all to continue? 
                 */
                ERROR_PRINT ("%s",
                             "Splashy could not determine FIFO path from your "
                             "configuration file. Setting to default path: "
                             "/etc/splashy/splashy.fifo");
                _spl_fifo_path = "/etc/splashy/splashy.fifo";
        }
        /*
         * 0 seconds and 10 nano seconds: 10 x 1x10^-9 
         */
        _sleep.tv_sec = 0;
        _sleep.tv_nsec = 10;

        FILE *spl_fifo = NULL;
        tv.tv_sec = timeout;
        tv.tv_usec = 0;

        FD_ZERO (&descriptors);
        FD_SET (pipe_fd, &descriptors);

        /*
         * set the cancellation parameters -- - Enable thread cancellation - 
         * Defer the action of the cancellation 
         */
        pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
        pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, NULL);

        /*
         * run our progress bar in a tight loop 
         */
        while (1)
        {
                if (exiting)
                {
                        sched_yield ();
                        nanosleep (&_sleep, NULL);
                        continue;
                }
                /*
                 * if we are in test mode, update progressbar continuously 
                 */
                if (_preview == TRUE)
                {
                        _sleep.tv_nsec = 1000;
                        /*
                         * since we are testing, allow progressbar to loop 
                         */
                        if (i >= 100)
                        {
                                i = 0;
                                forward_flag = (forward_flag) ? FALSE : TRUE;
                                /*
                                 * if (forward_flag)
                                 * g_printerr("Forward...\n"); else
                                 * g_printerr("Backward...\n");
                                 */

                                video_reset_progressbar_counters ();
                                video_set_progressbar_forward (forward_flag);
                                /*
                                 * re-draw the image so that we can get our
                                 * progressbar filled 
                                 */
                                video_change_splash (&video,
                                                     _current_background,
                                                     _show_progress);
                                sched_yield ();
                                nanosleep (&_sleep, NULL);
                        }
                        g_printerr ("Updating progressbar by %d+10\n", i);
                        video_update_progressbar (&video, &video.progressbar,
                                                  i += 10);
                        sched_yield ();
                        nanosleep (&_sleep, NULL);
                }
                else
                {
                        /*
                         * if no special keys were pressed, read from our
                         * fifo and continue waiting for events 
                         */
                        close (pipe_fd);
                        pipe_fd =
                                open (_spl_fifo_path, O_RDONLY | O_NONBLOCK);

                        if (timeout != 0)
                        {
                                tv.tv_sec = timeout;
                                err = select (pipe_fd + 1, &descriptors, NULL,
                                              NULL, &tv);
                        }
                        else
                        {
                                err = select (pipe_fd + 1, &descriptors, NULL,
                                              NULL, NULL);
                        }
                        if (err == -1)
                        {
                                /*
                                 * return;
                                 */
                                DEBUG_PRINT
                                        ("Error while selecting poll events for the Splashy FIFO %s",
                                         _spl_fifo_path);
                                splashy_child_exit ();
                        }

                        if (err == 0)
                        {
                                /*
                                 * Timeout 
                                 */
                                DEBUG_PRINT
                                        ("Timeout occurred when reading the Splashy FIFO %s",
                                         _spl_fifo_path);
                                splashy_child_exit ();
                        }

                        spl_fifo = fdopen (pipe_fd, "r");
                        if (!spl_fifo)
                        {
                                if (errno == EINTR)
                                {
                                        /*
                                         * no events in FIFO yet? release
                                         * some CPU cycles 
                                         */
                                        nanosleep (&_sleep, NULL);
                                        /*
                                         * continue; 
                                         */
                                        goto out;
                                }
                                DEBUG_PRINT
                                        ("Can't open the Splashy FIFO %s for reading",
                                         _spl_fifo_path);
                                /*
                                 * instead of exiting sleep for a while and
                                 * loop back the reason is because in a
                                 * initramfs we can't read from the fifo at
                                 * early boot... 
                                 */
                                _sleep.tv_sec += 1;     /* add one seconds to 
                                                         * sleep */
                                nanosleep (&_sleep, NULL);
                                /*
                                 * continue; 
                                 */
                                goto out;
                        }
                        /*
                         * fifo is read and a 'handler' takes care of the
                         * rest 
                         */
                        spl_read_fifo (spl_fifo);
                        /*
                         * note: spl_read_fifo will close spl_fifo when
                         * done... that in turn closes pipe_fd. We re-close
                         * pipe_fd anyway... 
                         */
                      out:
                        close (pipe_fd);
                }
                pthread_testcancel ();
                sched_yield ();
                nanosleep (&_sleep, NULL);
        }                       /* ends while(1) */
        pthread_exit (NULL);    /* never reached */
}
inline void *
keyevent_loop (void *data)
{
        /*
         * this function needs to always handle keyevents and not yield cpu too often
         */

        /*
         * counter to know when to yield CPU cycles * to other threads 
         */
        guint i_cpu = 0;

        /*
         * setup an event interface 
         */
        DFBInputEvent event;

        /*
         * set the cancellation parameters -- - Enable thread cancellation - 
         * Defer the action of the cancellation 
         */
        pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
        pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, NULL);

        while (1)
        {
                if (exiting)
                {
                        sched_yield ();
                        continue;
                }
                /*
                 * sub-parent (we are a fork after all). init is our parent 
                 */
                DFBCHECK (video.ev_buffer->WaitForEvent (video.ev_buffer));
                /*
                 * get all events from our event buffer (fifo style) and
                 * process them. 
                 */
                while (video.ev_buffer->HasEvent (video.ev_buffer) == DFB_OK
                       && video.ev_buffer->GetEvent (video.ev_buffer,
                                                     DFB_EVENT (&event)) ==
                       DFB_OK)
                {
                        /*
                         * exit if ESC or F2 is pressed also kill our
                         * sub_child while at it 
                         */
                        if (event.type == DIET_KEYPRESS)
                        {
                                if (event.key_id == DIKI_ESCAPE)
                                {
                                        splashy_child_exit ();
                                }
                                else if (event.key_id == DIKI_F2)
                                {
                                        /*
                                         * when F2 is pressed we display a
                                         * small theme-defined surface where 
                                         * text scrolls showing the currect
                                         * /dev/vcs[1,2,3,4,5,6,7] text 
                                         * @see verbose_text_loop
                                         */
                                        if (F2_toggle_pressed != TRUE)
                                        {
                                                F2_toggle_pressed = TRUE;
                                                show_textbox_area = TRUE;
                                                /* now show the box */
                                                video_printline_s (&video,
                                                                "",_show_progress);
                                        }
                                        else
                                        {
                                                F2_toggle_pressed = FALSE;
                                                show_textbox_area = FALSE;
                                                /*
                                                 * pressing F2 also clears the 
                                                 * switched_to_verbose flag
                                                 * @see verbose_text_loop
                                                 * we need to re-draw the screen
                                                 */
                                                video_change_splash (&video,
                                                                     _current_background,
                                                                     _show_progress);
                                        }
                                }
                        }
                }               /* ends while ev_buffer */
                /*
                 * every 100 tries check to see if the thread has been
                 * cancelled if the thread has not been cancelled then yield 
                 * the thread's LWP to another thread that may be able to run 
                 */
                if (i_cpu % 100 == 0)
                {
                        pthread_testcancel ();
                        sched_yield ();
                }
                i_cpu++;
        }                       /* end while(1) */
        pthread_exit (NULL);    /* never reached */
}

/**
 * @desc _splashy_child handles both start and stop routines for splashy during boot and shutdown respectively
 * Note that the fork() happens before this function is called by its
 * wrappers: splashy_child_start() and splashy_child_stop()
 * @see splashy_child_start() and @see splashy_child_stop()
 * @param seq sequence command to execute: "boot" or "shutdown"
 * @param show_progress whether we should display a progress bar
 * @return
 */
gint
_splashy_child (const gchar * seq, gboolean show_progress)
{
        gint i;                 /* general purpose counter */
        gint thr_id_c, thr_id_d;        /* thread ID for the newly * created
                                         * thread: c - * keyboard events d -
                                         * FIFO * handler */
        pthread_t p_thread_c, p_thread_d;       /* we need thread C * and D
                                                 * to manage * our keyboard * 
                                                 * events and FIFO */

        gint thr_id_f;          /* Handles /dev/vcs1 text to be displayed to
                                 * splashy overlay textbox ... when pressing
                                 * F2 we launch a thread to handle the
                                 * console text */
        pthread_t p_thread_f;

        const gchar *background = NULL;
        gboolean _preview = FALSE;

        /*
         * handle exit signals
         * man 7 signal
         */
        /*
         * (void) signal (SIGUSR1, sig_progress_update); (void) signal
         * (SIGUSR2, sig_progress_update);
         */
        (void) signal (SIGHUP, sig_hup_handler);
        (void) signal (SIGTERM, SIG_IGN);
        (void) signal (SIGINT, sig_exit_safe);
        (void) signal (SIGQUIT, sig_exit_safe);
        /*
         * TODO handling SIGABRT causes
         * "libgcc_s.so.1 must be installed for pthread_cancel to work" ?
         */
        (void) signal (SIGABRT, sig_exit_safe);

        if (g_ascii_strncasecmp (seq, "boot", 4) == 0)
        {
                background =
                        _splashy_theme_image_path (SPL_CURRENT_THEME,
                                                   "/splashy/background/boot");
                /*
                 * fallback to default theme if the current theme has no
                 * image defined 
                 */
                if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
                        background =
                                _splashy_theme_image_path (SPL_DEFAULT_THEME,
                                                           "/splashy/background/boot");
        }
        else if (g_ascii_strncasecmp (seq, "shutdown", 8) == 0)
        {
                background =
                        _splashy_theme_image_path (SPL_CURRENT_THEME,
                                                   "/splashy/background/shutdown");
                if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
                        background =
                                _splashy_theme_image_path (SPL_DEFAULT_THEME,
                                                           "/splashy/background/shutdown");
        }
        else if (g_ascii_strncasecmp (seq, "test", 4) == 0
                 || g_ascii_strncasecmp (seq, "preview", 7) == 0)
        {
                /*
                 * allows preview of Splashy current theme and progressbar 
                 */
                _preview = TRUE;
                /*
                 * set bg image to boot 
                 */
                background =
                        _splashy_theme_image_path (SPL_CURRENT_THEME,
                                                   "/splashy/background/boot");
                /*
                 * fallback to default theme if the current theme has no
                 * image defined 
                 */
                if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
                        background =
                                _splashy_theme_image_path (SPL_DEFAULT_THEME,
                                                           "/splashy/background/boot");
        }
        else
        {
                background =
                        _splashy_theme_image_path (SPL_CURRENT_THEME,
                                                   "/splashy/background/errorimg");
                if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
                        background =
                                _splashy_theme_image_path (SPL_DEFAULT_THEME,
                                                           "/splashy/background/errorimg");
        }
        DEBUG_PRINT ("_splashy_child() starting splash with image <<%s>>",
                     background);
        if (!g_file_test (background, G_FILE_TEST_IS_REGULAR))
        {
                ERROR_PRINT ("Fatal: background image not found <<%s>>",
                             background);
                splashy_child_exit ();
        }

        /*
         * close all open filehandles
         */
        for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
        {
                if (i == 2)
                        continue;       /* never close stderr */
                DEBUG_PRINT ("Filehandle closed %d", i);
                close (i);
        }

        video_start_splash (&video, background);
        _current_background = background;       /* safe for future reference */

        if (show_progress == TRUE)
        {
                /*
                 * starting bars stuff! 
                 */
                DEBUG_PRINT ("_splashy_child() starting progressbar %s", "");
                video_draw_progressbar (&video, &video.progressbar);
        }

        /*
         * Initialize optional text area 
         *
         * TODO if textbox is not enabled and autoverboseonerror is not enabled either
         * perhaps we don't even need to initialize the text area.
         * Note that this won't actually show the text area until we print
         * to it. @see verbose_text_loop() and the create thread below
         */
        video_start_text_area (&video, show_progress);

        /*
         * setup our main input: keyboard 
         */
        /*
         * TODO would we leak if we don't check if video.keyboard is not NULL 
         * ? We should make sure splashy can't be launched twice from this
         * same thread 
         */
        DFBCHECK (video.dfb->
                  GetInputDevice (video.dfb, DIDID_KEYBOARD,
                                  &video.keyboard));
        /*
         * create event buffer for the keyboard: to listen for events 
         */
        DFBCHECK (video.keyboard->
                  CreateEventBuffer (video.keyboard, &video.ev_buffer));
        /*
         * here is where the magic happens: because WaitForEvent*() waits idle
         * for events, we need to fork and handle this in separate threads
         */

        /*
         * thread C (main)
         * takes care of reading the FIFO and updating splashy internals 
         * according to the commands read from there
         */
        thr_id_c =
                pthread_create (&p_thread_c, NULL, fifo_loop,
                                (void *) &_preview);

        /*
         * thread D 
         * takes care of "listening" for keyboard events. For now:
         * ESC - exits all splashy threads
         * F2 - puts splashy in verbose mode
         */
        thr_id_d = pthread_create (&p_thread_d, NULL, keyevent_loop, NULL);

        thr_id_f =
                pthread_create (&p_thread_f, NULL, verbose_text_loop, NULL);

        /*
         * Wait till thread is finished reading the fifo... ie 'forever' 
         */
        pthread_join (p_thread_c, NULL);

        return 0;               /* never reached */
}                               /* end _splashy_child() */

/**
 * @desc the preferred way to exit splashy:
 */
void
splashy_child_exit ()
{
        cmd_exit (NULL);
}

/**
 * @desc what to do when splashy is starting up (boot process)
 * TODO
 * splashy child start/stop functions should allow argumenst from main.c
 */
void
splashy_child_start ()
{
        DEBUG_PRINT ("Entering splashy_child_start %s", "");
        const gchar *show_progress =
                xml_parser_get_text ("/splashy/progressbaronboot");
        _show_progress =
                (g_ascii_strncasecmp (show_progress, "yes", 3) ==
                 0) ? TRUE : FALSE;

        /*
         * TODO make this configurable from config.xml for e/a theme 
         */
        video_set_progressbar_forward (TRUE);

        gint ret = _splashy_child ("boot", _show_progress);
        if (ret < 0)
                ERROR_PRINT ("Could not start Splashy. %d\n", ret);
        DEBUG_PRINT ("Exiting splashy_child_start %d", ret);
        splashy_child_exit ();
}

/**
 * @desc what to do when splashy is in the stopping process (shutdown)
 */
void
splashy_child_stop ()
{
        const gchar *show_progress =
                xml_parser_get_text ("/splashy/progressbaronshutdown");
        _show_progress =
                (g_ascii_strncasecmp (show_progress, "yes", 3) ==
                 0) ? TRUE : FALSE;

        /*
         * TODO make this configurable from config.xml for e/a theme 
         */
        video_set_progressbar_forward (FALSE);

        gint ret = _splashy_child ("shutdown", _show_progress);
        if (ret < 0)
                g_print ("ERROR: Main process exited with %d\n", ret);
        splashy_child_exit ();
}

/**
 * @desc testing splashy
 */
void
splashy_child_test ()
{
        DEBUG_PRINT ("Entering splashy_child_test %s", "");
        const gchar *show_progress =
                xml_parser_get_text ("/splashy/progressbaronboot");
        _show_progress =
                (g_ascii_strncasecmp (show_progress, "yes", 3) ==
                 0) ? TRUE : FALSE;
        gint ret = _splashy_child ("test", _show_progress);
        if (ret < 0)
                ERROR_PRINT ("Could not start Splashy on test mode. %d\n",
                             ret);
        DEBUG_PRINT ("Exiting splashy_child_test %d", ret);
        splashy_child_exit ();
}

/*
 * TODO we might need to mount /proc 
 */
/*
 * #ifdef __linux__ / * Read /proc/cmdline. / stat("/", &st); if
 * (stat("/proc", &st2) < 0) { perror("Splashy: /proc"); return -1; } if
 * (st.st_dev == st2.st_dev) { if (mount("proc", "/proc", "proc", 0, NULL) <
 * 0) { perror("Splashy: mount /proc"); return -1; } didmount = 1; }
 * 
 * n = 0; r = -1; if ((fd = open("/proc/cmdline", O_RDONLY)) < 0) {
 * perror("Splashy: /proc/cmdline"); } else { buf[0] = 0; if ((n = read(fd,
 * buf, sizeof(buf) - 1)) >= 0) r = 0; else perror("Splashy: /proc/cmdline");
 * close(fd); } if (didmount) umount("/proc");
 * 
 * if (r < 0) return r;
 * 
 * #endif 
 */
