/************************************************************************
 *                      splashy_video.c                                 *
 *                                                                      *
 *  2005-05-28 03:01 EDT                                                *
 *  Copyright  2005  Otavio Salvador <otavio@debian.org>                *
 *                   Luis Mondesi    <lemsx1@gmail.com>                 *
 ************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>             /* abs() */

#include <stdio.h>
#include <string.h>
#include <time.h>               /* nanosleep() */
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/fb.h>
/*
 * 64bit arch fix bug#302647 in alioth and
 * bug#354856 on bugs.debian.org/splashy (DBTS)
 */
#include <directfb.h>

#include "common_macros.h"
#include "common_functions.h"
#include "splashy_functions.h"
#include "xml_parser.h"
#include "xml_config.h"
#include "splashy_video.h"

static const gchar *_current_background = NULL;
static gchar *fb_dev_name = NULL;       /* such as /dev/fb0 */
static gint fb_dev_fd;          /* handle for fb_dev_name */
static struct fb_var_screeninfo fb_vinfo;       /* variable info */
static struct fb_var_screeninfo fb_orig_vinfo;  /* variable info to restore
                                                 * later */
static gint fb_bpp;             /* 32: 32 24: 24 16: 16 15: 15 */

static IDirectFBFont *font = NULL;
static guint last_text_y_position = 0;
static guint _last_progress = 0;
static guint _last_w = 0;
static guint _last_x = 0;

static gboolean draw_progressbar_forward = TRUE;

/*
 * helper functions 
 */

/*
 * be very carefull when calling this function 
 * as it's not thread safe
 */
void
video_reset_progressbar_counters ()
{
        _last_progress = 0;
        _last_w = 0;
        _last_x = 0;
}

void
video_set_progressbar_forward (gboolean go_forward)
{
        draw_progressbar_forward = go_forward;
}

static gint
fb_preinit (gint reset)
{
        static gint fb_preinit_done = 0;
        static gint fb_err = -1;

        if (reset)
        {
                fb_preinit_done = 0;
                return 0;
        }

        if (fb_preinit_done)
                return fb_err;
        fb_preinit_done = 1;

        if (!fb_dev_name && !(fb_dev_name = getenv ("FRAMEBUFFER")))
                fb_dev_name = g_strdup ("/dev/fb0");

        DEBUG_PRINT ("Using device %s\n", fb_dev_name);

        if ((fb_dev_fd = open (fb_dev_name, O_RDWR)) == -1)
        {
                ERROR_PRINT ("Can't open %s", fb_dev_name);
                ERROR_PRINT ("%s\n", strerror (errno));
                goto err_out;
        }
        if (ioctl (fb_dev_fd, FBIOGET_VSCREENINFO, &fb_vinfo))
        {
                ERROR_PRINT ("[fbdev2] Can't get VSCREENINFO: %s\n",
                             strerror (errno));
                goto err_out;
        }
        fb_orig_vinfo = fb_vinfo;

        fb_bpp = fb_vinfo.bits_per_pixel;

        /*
         * 16 and 15 bpp is reported as 16 bpp 
         */
        if (fb_bpp == 16)
                fb_bpp = fb_vinfo.red.length + fb_vinfo.green.length +
                        fb_vinfo.blue.length;

        fb_err = 0;
        return 0;
      err_out:
        if (fb_dev_fd >= 0)
                close (fb_dev_fd);
        fb_dev_fd = -1;
        fb_err = -1;
        return -1;
}

static guint
preinit (const char *subdevice)
{
        if (subdevice)
        {
                if (fb_dev_name)
                        free (fb_dev_name);
                fb_dev_name = strdup (subdevice);
        }
        return fb_preinit (0);
}

/*
 * Public members 
 */
void
video_allow_vt_switching ()
{
        DEBUG_PRINT ("video_allow_vt_switching(%s) called", "");
        /*
         * Please read functions.c about why NOT to set vt-switching here.
         * 2005-04-28 20:42 EDT - Luis Mondesi <lemsx1@gmail.com> 
         */
        /*
         * DirectFBSetOption( "vt-switching", NULL); 
         */
        /* DirectFBSetOption ("force-windowed", NULL); */
        DEBUG_PRINT ("video_allow_vt_switching(%s) exited", "");
}

void
video_get_divider (splashy_video_t * video, gint * divider_w,
                   gint * divider_h)
{
        /*
         * Source image width and height 
         */
        gint source_width, source_height;

        /*
         * If a source image width and height is specified, assume
         * pixel-precise placement of progress bar rather than percentage. 
         */
        source_width =
                xml_parser_get_int ("/splashy/background/dimension/width",
                                    10);
        source_height =
                xml_parser_get_int ("/splashy/background/dimension/height",
                                    10);
        if (source_width > 0 && source_height > 0)
        {
                /*
                 * The value is based on specified units 
                 */
                *divider_w = source_width;
                *divider_h = source_height;
        }
        else
        {
                /*
                 * The value is based on percentage 
                 */
                *divider_w = *divider_h = 100;
        }
}

gint
video_draw_progressbar (splashy_video_t * video, DFBRectangle * progressbar)
{
        gint screen_width, screen_height;
        gint divider_w, divider_h;
        guint rectangle_red, rectangle_green, rectangle_blue, rectangle_alpha;

        const gchar *draw_progress =
                xml_parser_get_text ("/splashy/progressbar/border/enable");
        gboolean draw_progress_border = (draw_progress
                                         &&
                                         g_ascii_strncasecmp (draw_progress,
                                                              "yes",
                                                              3) ==
                                         0) ? TRUE : FALSE;
        DEBUG_PRINT ("Printing progress border: %d", draw_progress_border);

        video_get_divider (video, &divider_w, &divider_h);

        video->primary_surface->GetSize (video->primary_surface,
                                         &screen_width, &screen_height);

        progressbar->x = screen_width *
                xml_parser_get_int ("/splashy/progressbar/dimension/x",
                                    10) / divider_w;
        progressbar->y =
                screen_height *
                xml_parser_get_int ("/splashy/progressbar/dimension/y",
                                    10) / divider_h;
        progressbar->w =
                screen_width *
                xml_parser_get_int ("/splashy/progressbar/dimension/width",
                                    10) / divider_w;
        progressbar->h =
                screen_height *
                xml_parser_get_int ("/splashy/progressbar/dimension/height",
                                    10) / divider_h;

        if (progressbar->x < 0 ||
            progressbar->y < 0 || progressbar->w < 0 || progressbar->h < 0)
        {
                g_printerr
                        ("ERROR: progress bar coordinates (x,y,width,height) had negative values. Please check that your configuration file exists and it's readable\n");
                return 1;
        }

        if (draw_progress_border == TRUE)
        {
                /*
                 * Check progress bar border colours. Assume we want a border 
                 * if they are all properly specified. 
                 */
                rectangle_red =
                        xml_parser_get_int
                        ("/splashy/progressbar/border/color/red", 10);
                rectangle_green =
                        xml_parser_get_int
                        ("/splashy/progressbar/border/color/green", 10);
                rectangle_blue =
                        xml_parser_get_int
                        ("/splashy/progressbar/border/color/blue", 10);
                rectangle_alpha =
                        xml_parser_get_int
                        ("/splashy/progressbar/border/color/alpha", 10);
                if (!
                    (rectangle_red > 255 || rectangle_green > 255
                     || rectangle_blue > 255 || rectangle_alpha > 255))
                {
                        DEBUG_PRINT ("Border draw: x = %d", progressbar->x);
                        DEBUG_PRINT ("Border draw: y = %d", progressbar->y);
                        DEBUG_PRINT ("Border draw: w = %d", progressbar->w);
                        DEBUG_PRINT ("Border draw: h = %d", progressbar->h);
                        DEBUG_PRINT ("Border color: red = %d", rectangle_red);
                        DEBUG_PRINT ("Border color: green = %d",
                                     rectangle_green);
                        DEBUG_PRINT ("Border color: blue = %d",
                                     rectangle_blue);
                        DEBUG_PRINT ("Border color: alpha = %d",
                                     rectangle_alpha);

                        video->primary_surface->SetColor (video->
                                                          primary_surface,
                                                          rectangle_red,
                                                          rectangle_green,
                                                          rectangle_blue,
                                                          rectangle_alpha);

                        /*
                         * Draw border so that it is one pixel around the bar 
                         */
                        video->primary_surface->DrawRectangle (video->
                                                               primary_surface,
                                                               progressbar->
                                                               x - 1,
                                                               progressbar->
                                                               y - 1,
                                                               progressbar->
                                                               w + 2,
                                                               progressbar->
                                                               h + 2);
                }
                else if (rectangle_red < 0 || rectangle_green < 0 ||
                         rectangle_blue < 0 || rectangle_alpha < 0)
                {
                        g_printerr
                                ("ERROR: Progress bar border colour not properly specified");
                }
                else
                {
                        DEBUG_PRINT ("No progress bar border specified%s",
                                     "");
                }
        }
        rectangle_red =
                xml_parser_get_int ("/splashy/progressbar/color/red", 10);
        rectangle_green =
                xml_parser_get_int ("/splashy/progressbar/color/green", 10);
        rectangle_blue =
                xml_parser_get_int ("/splashy/progressbar/color/blue", 10);
        rectangle_alpha =
                xml_parser_get_int ("/splashy/progressbar/color/alpha", 10);

        /*
         * set the filling color of the progressbar to be 
         * used by video_update_progress()
         */
        DEBUG_PRINT ("Draw red   = %d", rectangle_red);
        DEBUG_PRINT ("Draw green = %d", rectangle_green);
        DEBUG_PRINT ("Draw blue  = %d", rectangle_blue);
        DEBUG_PRINT ("Draw alpha = %d", rectangle_alpha);
        if (rectangle_red > 255 ||
            rectangle_green > 255 ||
            rectangle_blue > 255 || rectangle_alpha > 255)
        {
                g_printerr
                        ("ERROR: Red, green, blue or alpha tags in the configuration file contain wrong values");
                return 3;
        }

        /*
         * TODO make this configurable from theme.xml
         * /splashy/progressbar/background/color/red...
         * for now picking gray
         */
        if (draw_progressbar_forward == TRUE)
        {
                /*
                 * fill the rectangle so that it's easier to do the reverse progress trick
                 * @see video_update_progress()
                 */
                video->primary_surface->SetColor (video->primary_surface,
                                                  127, 127, 127, 255);
        }
        else
        {
                video->primary_surface->SetColor (video->primary_surface,
                                                  rectangle_red,
                                                  rectangle_green,
                                                  rectangle_blue,
                                                  rectangle_alpha);
        }
        video->primary_surface->FillRectangle (video->primary_surface,
                                               progressbar->x,
                                               progressbar->y,
                                               progressbar->w,
                                               progressbar->h);
        /*
         * now invert the colors set 
         */
        if (draw_progressbar_forward != TRUE)
        {
                video->primary_surface->SetColor (video->primary_surface,
                                                  127, 127, 127, 255);
        }
        else
        {
                video->primary_surface->SetColor (video->primary_surface,
                                                  rectangle_red,
                                                  rectangle_green,
                                                  rectangle_blue,
                                                  rectangle_alpha);
        }

        /*
         * finally, reset our counters to their proper values
         * for forward and backward progress
         */
        _last_x = progressbar->w;       /* we go from back to front */
        _last_w = 0;            /* we start at progressbar->x really. 0 is
                                 * fine */

/*        IDirectFBSurface *_layer_surface;

        DFBCHECK (video->primary_layer->
                  GetSurface (video->primary_layer,
                              &_layer_surface));

        _layer_surface->Blit(_layer_surface,
                        video->primary_surface,
                        NULL, 0, 0);*/

        return 0;
}

/*
 * TODO 
 * if we get a perc that's less that the previus number
 * update the progressbar in reverse
 */
gint
video_update_progressbar (splashy_video_t * video,
                          DFBRectangle * progressbar, gint perc)
{
        /*
         * how many nanoseconds we sleep while updating the progress bar 
         */
        struct timespec _progress_sleep;
        guint i = 0, _w = 0;

        DEBUG_PRINT ("video_update_progressbar(%d) called", perc);
        /*
         * sanity check 
         */
        if (perc < 0)
                return 1;
        if (video == NULL)
                return 1;
        if (perc > 100)
                return 0;
        DEBUG_PRINT ("video_update_progressbar(%d) sanity checks passed",
                     perc);
        DEBUG_PRINT ("Progress x = %d", progressbar->x);
        DEBUG_PRINT ("Progress y = %d", progressbar->y);
        DEBUG_PRINT ("Progress w = %d", progressbar->w);
        DEBUG_PRINT ("Progress h = %d", progressbar->h);

        _progress_sleep.tv_sec = 0;
        _progress_sleep.tv_nsec = 100;

        _w = progressbar->w * perc / 100;
        if (_w <= 0)
        {
                DEBUG_PRINT ("Skipped progress width %d", _w);
                return 1;
        }

        if (draw_progressbar_forward != TRUE)
        {
                if (perc == 0)
                        _last_x = progressbar->w;

                guint _x = progressbar->x + (progressbar->w - _w);

                /*
                 * draw (smooth) progress backward 
                 */
                for (i = _last_x; i > _x; i--)
                {
                        video->primary_surface->FillRectangle (video->
                                                               primary_surface,
                                                               i,
                                                               progressbar->y,
                                                               progressbar->
                                                               w - (i -
                                                                    progressbar->
                                                                    x),
                                                               progressbar->
                                                               h);
                        _last_progress = i;
                        nanosleep (&_progress_sleep, NULL);
                }
                _last_x = _x;
        }
        else
        {
                /*
                 * draw (smooth) progress forward 
                 */
                for (i = _last_w; i <= _w; i++)
                {
                        video->primary_surface->FillRectangle (video->
                                                               primary_surface,
                                                               progressbar->x,
                                                               progressbar->y,
                                                               i,
                                                               progressbar->
                                                               h);
                        _last_progress = i;
                        nanosleep (&_progress_sleep, NULL);
                }
                _last_w = _w;
        }
        _last_progress = perc;
        return 0;
}

/**
 * updates the progress bar without smoothing it out
 */
gint
video_update_progressbar_quick (splashy_video_t * video,
                                DFBRectangle * progressbar, gint perc)
{
        guint _w = 0;

        DEBUG_PRINT ("video_update_progressbar_quick(%d) called", perc);
        /*
         * sanity check 
         */
        if (perc < 0)
                return 1;
        if (video == NULL)
                return 1;
        if (perc > 100)
                return 0;
        DEBUG_PRINT
                ("video_update_progressbar_quick(%d) sanity checks passed",
                 perc);
        DEBUG_PRINT ("Progress x = %d", progressbar->x);
        DEBUG_PRINT ("Progress y = %d", progressbar->y);
        DEBUG_PRINT ("Progress w = %d", progressbar->w);
        DEBUG_PRINT ("Progress h = %d", progressbar->h);

        _w = progressbar->w * perc / 100;
        if (_w <= 0)
        {
                DEBUG_PRINT ("Skipped progress width %d", _w);
                return 1;
        }

        DEBUG_PRINT ("Progress width %d", _w);

        if (draw_progressbar_forward != TRUE)
        {
                /*
                 * guint _x = progressbar->x + (progressbar->w - _w); 
                 */
                video->primary_surface->FillRectangle (video->
                                                       primary_surface,
                                                       _w,
                                                       progressbar->y,
                                                       progressbar->w - (_w -
                                                                         progressbar->
                                                                         x),
                                                       progressbar->h);
                _last_x = progressbar->x + (progressbar->w - _w);
        }
        else
        {
                video->primary_surface->FillRectangle (video->
                                                       primary_surface,
                                                       progressbar->x,
                                                       progressbar->y,
                                                       _w, progressbar->h);
                _last_w = _w;
        }

        _last_progress = perc;
        return 0;
}

static DFBEnumerationResult
video_modes_callback (gint width, gint height, gint bpp, void *data)
{
        splashy_videomode_t *video = (splashy_videomode_t *) data;
        gint overx = 0, overy = 0, closer = 0, over = 0;
        gint we_are_under = 0;

#ifdef DEBUG
        gchar *msg =
                g_strdup_printf
                ("video_modes_callback: Validator entered %i %i %i\n", width,
                 height, bpp);
        DEBUG_PRINT ("%s", msg);
        g_free (msg);
#endif
        overx = width - video->out_width;
        overy = height - video->out_height;
        if (!video->width)
        {
                video->width = width;
                video->height = height;
                video->overx = overx;
                video->overy = overy;
#ifdef DEBUG
                gchar *msg =
                        g_strdup_printf
                        ("video_modes_callback: Mode added %i %i %i\n", width,
                         height, bpp);
                DEBUG_PRINT ("%s", msg);
                g_free (msg);
#endif
        }
        if ((video->overy < 0) || (video->overx < 0))
                we_are_under = 1;       /* stored mode is smaller than req
                                         * mode */
        if (abs (overx * overy) < abs (video->overx * video->overy))
                closer = 1;     /* current mode is closer to desired res */
        if ((overx >= 0) && (overy >= 0))
                over = 1;       /* current mode is bigger or equaul to
                                 * desired res */
        if ((closer && (over || we_are_under)) || (we_are_under && over))
        {
                video->width = width;
                video->height = height;
                video->overx = overx;
                video->overy = overy;
#ifdef DEBUG
                gchar *msg =
                        g_strdup_printf
                        ("video_modes_callback: Better mode added %i %i %i\n",
                         width, height, bpp);
                DEBUG_PRINT ("%s", msg);
                g_free (msg);
#endif
        };
        return DFENUM_OK;
}

void
video_set_mode (splashy_video_t * video)
{
        /*
         * used to disable inputs from mouse and reduce overhead (about 15%
         * of CPU usage)
         */
        /*
         * FIXME video->dfb->SetCooperativeLevel (video->dfb,
         * DFSCL_FULLSCREEN); 
         */
        video->primary_layer->SetCooperativeLevel (video->primary_layer,
                                                   DLSCL_EXCLUSIVE);
        video->primary_layer->GetConfiguration (video->primary_layer,
                                                &video->primary_layer_config);

        DFBResult ret =
                video->dfb->EnumVideoModes (video->dfb, video_modes_callback,
                                            video->mode);
        if (ret)
        {
                DEBUG_PRINT
                        ("%s: Error while detecting full screen video modes",
                         "video->dfb->EnumVideoModes");
        }
        /*
         * http://directfb.org/docs/DirectFB_Reference/types.html#DFBSurfacePixelFormat 
         */

        /*
         * DLCONF_WIDTH|DLCONF_HEIGHT|DLCONF_PIXELFORMAT|DLCONF_SURFACE_CAPS 
         */
        video->primary_layer_config.width = video->mode->width;
        video->primary_layer_config.height = video->mode->height;
        /*
         * FIXME video->primary_layer_config.pixelformat = DSPF_ARGB; 
         */
        /*
         * DLOP_ALPHACHANNEL|DLOP_OPACITY 
         */

        ret = video->primary_layer->SetConfiguration (video->primary_layer,
                                                      &video->
                                                      primary_layer_config);

        DEBUG_PRINT ("Set resolution for %s:", "video->dfb");
        DEBUG_PRINT ("Width: %d", video->mode->width);
        DEBUG_PRINT ("Height: %d", video->mode->height);
}

void
video_start_splash (splashy_video_t * video, const gchar * background)
{
        DFBSurfaceDescription desc;
        DFBWindowDescription win_desc;

        /*
         * initializing Directfb
         */
        /*
         * - no-debug           := suppresses debug messages
         * - quiet              := suppresses all messages (but debug)
         * - graphics-vt        := puts directfb vt in graphics mode
         * - no-cursor          := disallow showing a cursor
         */

        _current_background = background;

        DFBCHECK (DirectFBInit (NULL, NULL));
        DirectFBSetOption ("quiet", NULL);
        DirectFBSetOption ("no-debug", NULL);
        DirectFBSetOption ("graphics-vt", NULL);
        DirectFBSetOption ("no-cursor", NULL);
        /* TODO doesn't solve anything: DirectFBSetOption ("dont-catch", "15"); */
        /* TODO too dangerous DirectFBSetOption ("block-all-signals", NULL); */


        DFBCHECK (DirectFBCreate (&video->dfb));

        video->mode = g_new0 (splashy_videomode_t, 1);
        /*
         * set our expectation to a very big number 
         */
        preinit (NULL);
        video->mode->out_height = fb_vinfo.yres;
        video->mode->out_width = fb_vinfo.xres;

        DEBUG_PRINT ("Setting min Width (x) resolution to %d",
                     video->mode->out_width);
        DEBUG_PRINT ("Setting min Height (y) resolution to %d",
                     video->mode->out_height);

        DFBCHECK (video->dfb->CreateImageProvider (video->dfb, background,
                                                   &video->provider));
        DFBCHECK (video->provider->GetSurfaceDescription (video->provider,
                                                          &desc));
        /*
         * flags to set the default surface as main surface
         */
        desc.flags = DSDESC_CAPS;
        desc.caps = DSCAPS_PRIMARY;

        /*
         * store our primary layer as this will be use for setting the opacity
         * levels later
         */
        DFBCHECK (video->dfb->GetDisplayLayer (video->dfb,
                                               DLID_PRIMARY,
                                               &video->primary_layer));

        /*
         * we need to call this after primary_layer was set. @see video_set_mode()
         */
        video_set_mode (video);

        /*
         * get our primary_surface, this will hold the progressbar, textbox
         * and others
         */
        DFBCHECK (video->primary_layer->
                  GetSurface (video->primary_layer,
                              &video->primary_surface));

        win_desc.flags = (DWDESC_POSX | DWDESC_POSY |
                          DWDESC_WIDTH | DWDESC_HEIGHT);
        win_desc.posx = 0;
        win_desc.posy = 0;
        win_desc.width = video->mode->out_width;
        win_desc.height = video->mode->out_height;
        /* TODO do we really need this? win_desc.caps = DWCAPS_ALPHACHANNEL; */

        DFBCHECK (video->primary_layer->CreateWindow (video->primary_layer,
                                                      &win_desc,
                                                      &video->
                                                      primary_window));
        
        DFBCHECK (video->primary_window->
                  GetSurface (video->primary_window,
                              &video->primary_window_surface));

        if (g_ascii_strncasecmp
            (xml_parser_get_text ("/splashy/fadein"), "yes", 3) == 0)
                video->primary_window->SetOpacity (video->primary_window,
                                                   0x0);

        /*
         * allow surface to have alpha channels 
         * fade in effect won't work with alpha channels in window
         * because we use Opacity levels. See windows capability DWCAPS_ALPHACHANNEL
         * above
         */
        /*
         * DFBCHECK (video->primary_surface->SetBlittingFlags
         * (video->primary_surface, DSBLIT_BLEND_ALPHACHANNEL));
         */

        /*
         * it writes on the framebuffer the background image
         */
        video->provider->RenderTo (video->provider, video->primary_window_surface,
                                   NULL);
        video->provider->RenderTo (video->provider, video->primary_surface,
                                   NULL);

        /*
         * fade in effect 
         */

        if (g_ascii_strncasecmp
            (xml_parser_get_text ("/splashy/fadein"), "yes", 3) == 0)
                video_fade_in (video);

}

void
video_change_splash (splashy_video_t * video,
                     const gchar * newimage, gboolean show_progress)
{
        video->dfb->CreateImageProvider (video->dfb, newimage,
                                         &video->provider);
        video->provider->RenderTo (video->provider, video->primary_surface,
                                   NULL);
        /*
         * restore progressbar 
         */
        if (show_progress == TRUE)
        {
                video_draw_progressbar (video, &video->progressbar);
                if (_last_progress >= 0)
                {
                        DEBUG_PRINT ("Restoring progressbar to %d ticks\n",
                                     _last_progress);
                        video_update_progressbar_quick (video,
                                                        &video->progressbar,
                                                        _last_progress);
                }
        }
        _current_background = newimage;
}

gint
get_time_in_ms (void)
{
        struct timeval tp;
        struct timezone tz;

        if (gettimeofday (&tp, &tz) == 0)
                return (tp.tv_sec * 1000) + (tp.tv_usec / 1000);
        return 0;               /* should never be reached */
}

/*
 * Fade the opacity from its current value to the specified.
 * helper for video_fade_in/out()
 */
static void
_window_fade (splashy_video_t * win, __u8 opacity, long ms_duration)
{
        long t1, t2;
        int diff;
        /*
         * IDirectFBDisplayLayer *layer = win->primary_layer;
         */
        IDirectFBWindow *window = win->primary_window;

        /*
         * Retrieve its current opacity. 
         */
        window->GetOpacity (window, &win->opacity);

        /*
         * TODO how do we know the current opacity level 
         */
        /*
         * Retrieve its current opacity. 
         */
        /*
         * layer->GetOpacity (layer, &win->opacity);
         */

        /*
         * Calculate the opacity difference. 
         */
        diff = opacity - win->opacity;
        /*
         * fprintf(stderr,"diff %d\n",diff); 
         */
        /*
         * Initialize both time stamps. 
         */
        t1 = t2 = get_time_in_ms ();
        /*
         * fprintf(stderr,"t1 %lu\n",t2); 
         */

        do
        {
                /*
                 * Calculate an intermediate opacity. 
                 */
                __u8 op = win->opacity + ((t2 - t1 + 1) * diff / ms_duration);
                /*
                 * fprintf(stderr,"op %d\n",op); 
                 */
                /*
                 * Set the intermediate opacity. 
                 */
                window->SetOpacity (window, op);
                /*
                 * layer->SetOpacity (layer, op); 
                 */

                /*
                 * Update the time stamp. 
                 */
                t2 = get_time_in_ms ();
                /*
                 * fprintf(stderr,"t2 %lu\n",t2); 
                 */
        }
        while (t2 - t1 < ms_duration);
}

/**
 * A simple function to create the effect of "fading in" for the
 * primary surface
 * @see splashy_child_boot() on splashy_functions.c
 * @param video main video holding our primary window
 */
void
video_fade_in (splashy_video_t * video)
{
        /*
         * fprintf(stderr,"video_fade_in called %s\n",""); 
         */

        IDirectFBWindow *window = video->primary_window;
        /*
         * start from almost black 
         */
        window->SetOpacity (window, 1);

        _window_fade (video, 255, 1000);

        /*
         * Set the exact opacity. 
         */
        window->SetOpacity (window, 255);
        video->opacity = 255;
}

/**
 * A simple function to create the effect of "fading out" for the
 * primary surface
 * @see splashy_child_exit() on splashy_functions.c
 * @param video main video holding our primary window
 */
void
video_fade_out (splashy_video_t * video)
{
        /*
         * fprintf(stderr,"video_fade_out called %s\n","");
         */

        IDirectFBWindow *window = video->primary_window;
        /*
         * start from fully opaque 
         */
        window->SetOpacity (window, 255);

        _window_fade (video, 0, 1000);

        /*
         * Set the exact opacity. 
         */
        window->SetOpacity (window, 0);
        video->opacity = 0;
}

/**
 * A helper function to clear up the offscreen surface
 */

void
video_clear_offscreen (splashy_video_t * video, gboolean show_progress)
{
        gint red, green, blue, alpha;
        if (video->textbox == NULL)
                return;
        /*
         * TODO later we will need to add textbox border support here: const
         * Get the tinting colour 
         */
        red = xml_parser_get_int ("/splashy/textbox/color/red", 10);
        green = xml_parser_get_int ("/splashy/textbox/color/green", 10);
        blue = xml_parser_get_int ("/splashy/textbox/color/blue", 10);
        alpha = xml_parser_get_int ("/splashy/textbox/color/alpha", 10);
        if (red < 0 || red > 255)
                red = 128;
        if (green < 0 || green > 255)
                green = 128;
        if (blue < 0 || blue > 255)
                blue = 128;
        if (alpha < 0 || alpha > 255)
                alpha = 255;

        /*
         * copy the "clean" surface from our primary surface to offscreen 
         */
        video_change_splash (video, _current_background, show_progress);

        /*
         * now we need to flip these surfaces so that the background tint
         * gets displayed correctly 
         */
        video->textbox->offscreen->Blit (video->textbox->offscreen,
                                         video->primary_surface,
                                         &video->textbox->area, 0, 0);

        video->textbox->offscreen->Blit (video->primary_surface,
                                         video->textbox->offscreen,
                                         &video->textbox->area, 0, 0);

        /*
         * Tint the box in the off-screen surface 
         */
        video->textbox->offscreen->SetDrawingFlags (video->textbox->offscreen,
                                                    DSDRAW_BLEND);
        video->textbox->offscreen->SetColor (video->textbox->offscreen, red,
                                             green, blue, alpha);
        video->textbox->offscreen->FillRectangle (video->textbox->offscreen,
                                                  0, 0,
                                                  video->textbox->area.w,
                                                  video->textbox->area.h);
}

/**
* Initialize and draw an alpha-transparent text box.
*
* @return
*/
void
video_start_text_area (splashy_video_t * video, gboolean show_progress)
{
        gint divider_w, divider_h;
        gint screen_width, screen_height;
        gint red, green, blue;  /* alpha; */
        gint temp;
        const gchar *fontface;

        /*
         * Check if there will be a text area. 
         */
        /*
         * TODO autoverboseonerror implies /splashy/textbox/enable="yes" 
         */
        const gchar *enable = xml_parser_get_text ("/splashy/textbox/enable");

        if (g_ascii_strncasecmp (enable, "yes", 3) != 0)
                return;

        video->primary_surface->GetSize (video->primary_surface,
                                         &screen_width, &screen_height);
        video_get_divider (video, &divider_w, &divider_h);

        video->textbox = g_new0 (splashy_textbox_t, 1);

        /*
         * Set some defaults so that there are no divide by zeroes if the
         * checks below fail 
         */
        video->textbox->area.x = video->textbox->area.w = divider_w / 4;
        video->textbox->area.y = video->textbox->area.h = divider_h / 4;

        /*
         * Read in the text box dimensions 
         */
        temp = xml_parser_get_int ("/splashy/textbox/dimension/x", 10);
        if (temp >= 0 && temp < divider_w)
                video->textbox->area.x = temp * screen_width / divider_w;

        temp = xml_parser_get_int ("/splashy/textbox/dimension/y", 10);
        if (temp >= 0 && temp < divider_h)
                video->textbox->area.y = temp * screen_height / divider_h;

        temp = xml_parser_get_int ("/splashy/textbox/dimension/width", 10);
        if (temp > 0 && temp <= divider_w)
                video->textbox->area.w = temp * screen_width / divider_w;

        temp = xml_parser_get_int ("/splashy/textbox/dimension/height", 10);
        if (temp > 0 && temp <= divider_h)
                video->textbox->area.h = temp * screen_height / divider_h;

        video->textbox->desc.width = video->textbox->area.w;
        video->textbox->desc.height = video->textbox->area.h;
        video->textbox->desc.caps = DSCAPS_SYSTEMONLY;  // Only in system
        // memory
        video->textbox->desc.flags =
                DSDESC_CAPS | DSDESC_WIDTH | DSDESC_HEIGHT;

        /*
         * Grab the textbox area to a seperate surface 
         */
        video->dfb->CreateSurface (video->dfb, &video->textbox->desc,
                                   &video->textbox->offscreen);
        video->textbox->offscreen->Blit (video->textbox->offscreen,
                                         video->primary_surface,
                                         &video->textbox->area, 0, 0);

        /*
         * clear the offscreen surface and set the drawing flags
         */
        video_clear_offscreen (video, show_progress);

        /*
         * Get the text colour (no alpha) 
         */
        red = xml_parser_get_int ("/splashy/textbox/text/color/red", 10);
        green = xml_parser_get_int ("/splashy/textbox/text/color/green", 10);
        blue = xml_parser_get_int ("/splashy/textbox/text/color/blue", 10);
        if (red < 0 || red > 255)
                red = 0;        /* if missing, make font color black */
        if (green < 0 || green > 255)
                green = 0;
        if (blue < 0 || blue > 255)
                blue = 0;

        /*
         * Establish the subsurface to which we print 
         */
        video->primary_surface->GetSubSurface (video->primary_surface,
                                               &video->textbox->area,
                                               &video->textbox->surface);
        video->textbox->surface->SetColor (video->textbox->surface, red,
                                           green, blue, 255);

        /*
         * Initialize the font 
         */
        video->textbox->fontdesc.flags = DFDESC_HEIGHT;
        fontface = xml_parser_get_text ("/splashy/textbox/text/font/file");
        temp = xml_parser_get_int ("/splashy/textbox/text/font/size", 10);
        video->textbox->fontdesc.height = temp * screen_height / divider_h;
        video->dfb->CreateFont (video->dfb, fontface,
                                &video->textbox->fontdesc, &font);
        if (font == NULL)
                video->dfb->CreateFont (video->dfb, NULL, NULL, &font);
        video->textbox->surface->SetFont (video->textbox->surface, font);

        /*
         * Set the position of the text line
         */
        last_text_y_position = abs (video->textbox->fontdesc.height);
}

/**
 * Print a line of text in the textbox.
 *
 * @return
 */
void
video_printline (splashy_video_t * video, gchar * string,
                 gboolean show_progress)
{
        if (video->textbox == NULL)
                return;

        /*
         * TODO 
         * we want to make sure (for now) that we don't make the textbox area
         * too dark. @see video_printline_s()
         */
        if (last_text_y_position > 0)
        {
                video_clear_offscreen (video, show_progress);
                last_text_y_position = 0;
        }

        /*
         * Copy the textbox background from off-screen surface 
         */
        video->textbox->offscreen->Blit (video->textbox->surface,
                                         video->textbox->offscreen,
                                         NULL, 0, 0);

        /*
         * Draw string to the clipped surface 
         */
        video->textbox->surface->DrawString (video->textbox->surface, string,
                                             -1, video->textbox->area.w / 2,
                                             (video->textbox->area.h / 2) +
                                             (video->textbox->fontdesc.
                                              height / 2),
                                             DSTF_CENTER | DSTF_BOTTOM);
}

/**
 * Print a line of text in the textbox and scroll down
 *
 * @return
 */
void
video_printline_s (splashy_video_t * video, gchar * string,
                   gboolean show_progress)
{
        if (video->textbox == NULL)
                return;

        /*
         * at the last line of the area minus one text line, we clear up 
         * our offscreen screen
         *
         * TODO 
         * we are double-copying this buffer, once from offscreen, once to
         * offscreen. which makes the background darker and darker...
         */
        if (last_text_y_position >=
            (video->textbox->area.h -
             abs (video->textbox->fontdesc.height * 2)))
        {
                video_clear_offscreen (video, show_progress);
                last_text_y_position = 0;
        }
        /*
         * Copy the textbox background from off-screen surface 
         */
        video->textbox->offscreen->Blit (video->textbox->surface,
                                         video->textbox->offscreen,
                                         NULL, 0, 0);

        last_text_y_position += abs (video->textbox->fontdesc.height * 2);

        /*
         * Draw string to the clipped surface 
         */
        video->textbox->surface->DrawString (video->textbox->surface, string,
                                             -1, 4,
                                             last_text_y_position, DSTF_LEFT);
        /*
         * Save the textbox background to off-screen surface 
         * Note, we clean, blend and save
         */
        video->textbox->offscreen->Blit (video->textbox->offscreen,
                                         video->primary_surface,
                                         &video->textbox->area, 0, 0);
        video->primary_surface->Blit (video->textbox->offscreen,
                                      video->textbox->surface, NULL, 0, 0);

}
